import os.path
import re
from pymzml.run import Reader

DATAFRAME_DATATYPES = {'experiment': str,
                       'ms_run': str,
                       'bins': str,
                       'proteins': str,
                       'peptide': str,
                       'peptide_mass': float,
                       'C': int,
                       'N': int,
                       'O': int,
                       'H': int,
                       'S': int,
                       'psm_id': int,
                       'psm_mz': float,
                       'psm_charge': int,
                       'psm_neutrons': float,
                       'psm_rank': int,
                       'psm_precursor_id': int,
                       'psm_precursor_mz': float,
                       'pattern_charge': int,
                       'pattern_precursor_id': int,
                       'pattern_total_intensity': float,
                       'pattern_peak_count': int,
                       'pattern_median_peak_spacing': float,
                       'pattern_mass_irregularity': float,
                       'ratio_na': float,
                       'ratio_fft': float,
                       'error_fft': float,
                       'error_clumpy': float,
                       'flag_peptide_contains_sulfur': bool,
                       'flag_peptide_has_modifications': bool,
                       'flag_peptide_assigned_to_multiple_bins': bool,
                       'flag_peptide_assigned_to_multiple_proteins': bool,
                       'flag_peptide_mass_and_elements_undefined': bool,
                       'flag_psm_has_low_confidence': bool,
                       'flag_psm_is_ambiguous': bool,
                       'flag_pattern_is_contaminated': bool,
                       'flag_pattern_is_wobbly': bool,
                       'flag_peak_at_minus_one_pos': bool}

DATAFRAME_COLUMNS = ['experiment', 'ms_run', 'bins', 'proteins', 'peptide', 'peptide_mass', 'C', 'N', 'O', 'H', 'S',
                     'psm_id', 'psm_mz', 'psm_charge', 'psm_neutrons', 'psm_rank', 'psm_precursor_id',
                     'psm_precursor_mz', 'pattern_charge', 'pattern_precursor_id', 'pattern_total_intensity',
                     'pattern_peak_count', 'pattern_median_peak_spacing', 'pattern_mass_irregularity',
                     'ratio_na', 'ratio_fft', 'error_fft', 'error_clumpy',
                     'flag_peptide_contains_sulfur', 'flag_peptide_has_modifications',
                     'flag_peptide_assigned_to_multiple_bins', 'flag_peptide_assigned_to_multiple_proteins',
                     'flag_peptide_mass_and_elements_undefined', 'flag_psm_has_low_confidence', 'flag_psm_is_ambiguous',
                     'flag_pattern_is_contaminated', 'flag_pattern_is_wobbly', 'flag_peak_at_minus_one_pos']

for i in range(20):  # intensities of the spectrum's peaks
    column_name = f'i{i}'
    DATAFRAME_COLUMNS.append(column_name)
    DATAFRAME_DATATYPES[column_name] = float

for m in range(20):  # masses of the spectrum's peaks
    column_name = f'm{m}'
    DATAFRAME_COLUMNS.append(column_name)
    DATAFRAME_DATATYPES[column_name] = float

for c in range(1, 7):  # clumpiness predictions
    column_name = f'c{c}'
    DATAFRAME_COLUMNS.append(column_name)
    DATAFRAME_DATATYPES[column_name] = float


def _is_mzid_file(file):
    return False


def _is_mzml_file(file):
    (file_name, extension) = os.path.splitext(file)
    return extension.lower() == '.mzml'


def _get_mzml_file_runid(file):
    parser: Reader = Reader(file)
    run_id = parser.info.get('run_id')
    parser.close()
    m = re.match('x[0-9]{4}_', run_id)
    if m:
        print(f'WARNING: run id of file {file} contains escaped special character "{m.group(0)}", consider using filenames '
              f'without special characters.')
    return run_id


def _is_target_spectrum_match_file(f):
    try:
        file = open(f)
        line = file.readline()
        outcome = 'Annotated Sequence' in line and 'PSM Ambiguity' in line and 'm/z [Da]' in line
        file.close()
        return outcome
    except FileNotFoundError:
        return False
    except UnicodeDecodeError:
        return False
    except IsADirectoryError:
        return False


class DataStore:
    def __init__(self):
        self.experiments = []
        self.ms_runs = []
        self.ms_run_address_book = {}

    def set_scope(self, peptide_file_stub, spectrum_file_stub):
        if os.path.isfile(peptide_file_stub):
            self.experiments.append(peptide_file_stub)
        elif os.path.isdir(peptide_file_stub):
            for f in os.listdir(peptide_file_stub):
                f = os.path.join(peptide_file_stub, f)
                if _is_target_spectrum_match_file(f) or _is_mzid_file(f):
                    self.experiments.append(f)
        else:
            raise FileNotFoundError(f'Could not read or find provided peptide file: {peptide_file_stub!r}')
        if os.path.isfile(spectrum_file_stub):
            self.ms_runs.append(spectrum_file_stub)
        elif os.path.isdir(spectrum_file_stub):
            for f in os.listdir(spectrum_file_stub):
                f = os.path.join(spectrum_file_stub, f)
                if _is_mzml_file(f):
                    self.ms_runs.append(f)
        for f in self.ms_runs:
            (file_name, extension) = os.path.splitext(f)
            self.ms_run_address_book[file_name] = f
            self.ms_run_address_book[os.path.basename(file_name)] = f
            self.ms_run_address_book[_get_mzml_file_runid(f)] = f
            self.ms_run_address_book[f] = f

    def get_ms_run_file_index(self, ms_run_file):
        return self.ms_runs.index(self.ms_run_address_book[ms_run_file])

    def get_experiment_index(self, experiment_file_name):
        return self.experiments.index(experiment_file_name)



import numpy as np
import re
from calisp import unimod

AMINOACID_COUNT: int = 20
AMINOACID_ELEMENT_COUNTS = np.array([  # C, N, O, H, S
                                    [2, 1, 2, 5, 0],  # G
                                    [5, 1, 2, 9, 0],  # P
                                    [3, 1, 2, 7, 0],  # A
                                    [5, 1, 2, 11, 0],  # V
                                    [6, 1, 2, 13, 0],  # L
                                    [6, 1, 2, 13, 0],  # I
                                    [5, 1, 2, 11, 1],  # M
                                    [3, 1, 2, 7, 1],  # C
                                    [9, 1, 2, 11, 0],  # F
                                    [9, 1, 3, 11, 0],  # Y
                                    [11, 2, 2, 12, 0],  # W
                                    [6, 3, 2, 9, 0],  # H
                                    [6, 2, 2, 14, 0],  # K
                                    [6, 4, 2, 14, 0],  # R
                                    [5, 2, 3, 10, 0],  # Q
                                    [4, 2, 3, 8, 0],  # N
                                    [5, 1, 4, 9, 0],  # E
                                    [4, 1, 4, 7, 0],  # D
                                    [3, 1, 3, 7, 0],  # S
                                    [4, 1, 3, 9, 0]], dtype=np.int64)   # T
WATER_ELEMENT_COUNTS = np.array([0, 0, 1, 2, 0], dtype=np.int64)
AMINOACID_IDS = 'G P A V L I M C F Y W H K R Q N E D S T'.split()
AMINOACID_INDEXES = {x: AMINOACID_IDS.index(x) for x in AMINOACID_IDS}
ELEMENT_MONOISOTOPIC_MASSES = np.array([12.0, 14.0030740048, 15.99491461956, 1.0078250322, 31.972070999],
                                       dtype=np.float64)
AMINOACID_NAMES = 'Glycine(G) Proline(P) Alanine(A) Valine(V) Leucine(L) Isoleucine(I) Methionine(M) Cysteine(C) ' \
                  'Phenylalanine(F) Tyrosine(Y) Tryptophan(W) Histidine(H) Lysine(K) Arginine(R) Glutamine(Q) ' \
                  'Asparagine(N) Glutamate(E) Aspartate(D) Serine(S) Threonine(T)'.split()
AMINOACID_INDEXED_NAMES = {x: AMINOACID_NAMES[AMINOACID_IDS.index(x)] for x in AMINOACID_IDS}
PROTON_MASS = 1.0072765
NEUTRON_MASS_SHIFT = 1.002

__ALL_UNIMOD_PEPTIDE_MODIFICATIONS_ELEMENT_COUNTS = []
__ALL_UNIMOD_PEPTIDE_MODIFICATIONS_EXTRA_NEUTRONS = []
__ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS = []


def load_unicode():
    for line in unimod.UNIMOD.splitlines():
        words = line.split("\t")
        if len(words) < 2:
            continue
        m = re.search('C([0-9+-]+)H([0-9+-]+)N([0-9+-]+)O([0-9+-]+)S([0-9+-]+)', words[0])
        if m:
            unimod_id = words[1].strip()
            element_counts = np.empty(5, dtype=np.int16)
            element_counts[0] = int(m.group(1))
            element_counts[1] = int(m.group(3))
            element_counts[2] = int(m.group(4))
            element_counts[3] = int(m.group(2))
            element_counts[4] = int(m.group(5))
            extra_neutrons = np.int16(0)
            __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_ELEMENT_COUNTS.append(element_counts)
            __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS.append(unimod_id)
            if len(words) > 2:
                extra_neutrons = np.int16(int(words[2]))
            __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_EXTRA_NEUTRONS.append(extra_neutrons)


def unimod_list():
    if not __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS:
        load_unicode()
    return __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS


def unimod_peptide_modifications_element_counts(i):
    if not __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS:
        load_unicode()
    return __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_ELEMENT_COUNTS[i]


def unimod_peptide_modifications_extra_neutrons(i):
    if not __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS:
        load_unicode()
    return __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_EXTRA_NEUTRONS[i]


def unimod_peptide_modifications_id(unimod_id):
    if not __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS:
        load_unicode()
    for i in range(len(__ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS)):
        if __ALL_UNIMOD_PEPTIDE_MODIFICATIONS_IDS[i] == unimod_id:
            return i
    return -1


def compute_peptide_mass(aminoacid_sequence: str, peptide_modifications, element_counts=np.zeros(5, dtype=np.int16)):
    for i in range(5):
        element_counts[i] += WATER_ELEMENT_COUNTS[i]
    for aa in aminoacid_sequence.upper():
        aa_element_counts = AMINOACID_ELEMENT_COUNTS[AMINOACID_INDEXES[aa]]
        for i in range(5):
            element_counts[i] += aa_element_counts[i]
            element_counts[i] -= WATER_ELEMENT_COUNTS[i]
    for modification in peptide_modifications:
        modification_element_counts = unimod_peptide_modifications_element_counts(modification)
        if modification_element_counts is None:
            raise Exception(f"unknown modification {modification} in {aminoacid_sequence} "
                            f"- molecular mass computation failed")
        for i in range(5):
            element_counts[i] += modification_element_counts[i]
    mass = np.float64(0)
    for i in range(5):
        mass += element_counts[i] * ELEMENT_MONOISOTOPIC_MASSES[i]
    return mass


def get_element_counts_at_pos(aminoacid_sequence: str, peptide_modifications, peptide_modification_positions, pos):
    aa = aminoacid_sequence[pos]
    if aa not in AMINOACID_INDEXES:
        raise Exception(f"unknown aminoacid {aa} in {aminoacid_sequence} - elements at position computation failed")
    aa_element_counts = AMINOACID_ELEMENT_COUNTS[AMINOACID_INDEXES[aa]]
    element_counts = np.zeros(5, dtype=np.int64)
    for i in range(5):
        element_counts[i] = aa_element_counts[i]
    # eliminate water from peptide bond, except at C-terminus
    if pos < len(aminoacid_sequence)-1:
        for i in range(5):
            element_counts[i] -= WATER_ELEMENT_COUNTS[i]
    for j in range(len(peptide_modifications)):
        if peptide_modification_positions[j] == pos:
            modification_element_counts = unimod_peptide_modifications_element_counts(peptide_modifications[j])
            if not modification_element_counts:
                raise Exception(f"unknown modification {peptide_modifications[j]} in {aminoacid_sequence} "
                                f"- molecular mass computation at pos {pos} failed")
            for i in range(5):
                element_counts[i] += modification_element_counts[i]
    return element_counts


def contains_sulfur(aminoacid_sequence):
    return 'C' in aminoacid_sequence or 'M' in aminoacid_sequence



import os
import re
import numpy as np
from calisp import element_count_and_mass_utils
from calisp import isotopic_pattern_utils

UNBINNED = 'unbinned'


class PeptideSpectrumMatchFileReader:
    def __init__(self, experiment_file_name, bin_delimiter='_'):
        self.experiment_file_name = experiment_file_name
        self.bin_delimiter = bin_delimiter
        self.unknown_modifications = set()
        self.parse_bins = bin_delimiter != '-'

    def __enter__(self):
        try:
            self.file = open(self.experiment_file_name)
            line = self.file.readline()
            line = line[1:-1]
            words = line.split("\"\t\"")
            self.columns = {'Annotated Sequence': -1,
                            'Confidence': -1,
                            'PSM Ambiguity': -1,
                            'Modifications': -1,
                            'First Scan': -1,
                            'Spectrum File': -1,
                            'Charge': -1,
                            'm/z [Da]': -1,
                            'Protein Accessions': -1}
            for k in self.columns.keys():
                self.columns[k] = words.index(k)
            return self
        except ValueError:
            raise Exception(f"Missing column in spectrum match file {self.experiment_file_name}")

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __iter__(self):
        return self

    def close(self):
        self.file.close()

    def __next__(self) -> {}:
        for line in self.file:
            # line = self.file.readline()
            line = line[1:-1]
            words = line.split("\"\t\"")
            psm_spectrum_id = int(words[self.columns['First Scan']])
            psm_is_above_threshold = words[self.columns['Confidence']] == 'High'
            psm_charge = int(words[self.columns['Charge']])
            psm_mz = float(words[self.columns['m/z [Da]']])
            psm_neutrons_in_ms2_spectrum = 0
            psm_is_ambiguous = 'unambiguous' != words[self.columns['PSM Ambiguity']].lower()

            peptide_aminoacid_sequence = words[self.columns['Annotated Sequence']].upper()
            peptide_modifications = []
            peptide_modification_positions = []
            ms_run_name = words[self.columns['Spectrum File']]
            (ms_run_name, extension) = os.path.splitext(ms_run_name)
            peptide_protein_names = set()
            peptide_bin_names = set()
            for protein_acc in words[self.columns['Protein Accessions']].replace("\t", " ").split(';'):
                protein_acc = protein_acc.strip()
                if self.parse_bins and self.bin_delimiter in protein_acc:
                    (bin_name, protein_name) = protein_acc.split(self.bin_delimiter, 1)
                else:
                    protein_name = protein_acc
                    bin_name = UNBINNED
                peptide_protein_names.add(protein_name)
                peptide_bin_names.add(bin_name)

            peptide_mass_undefined = False
            for modification_definition in words[self.columns['Modifications']].split(';'):
                modification_definition = modification_definition.strip()
                if not modification_definition:
                    continue
                m = re.match('[A-Z]([0-9]+)[(](.+?)[)]', modification_definition)
                if not m:
                    m = re.match('([CN])-Term(?:[(]Prot[)])*[(](.+?)[)]', modification_definition)
                if not m:
                    peptide_mass_undefined = True
                    self.unknown_modifications.add(modification_definition)
                    continue
                i = element_count_and_mass_utils.unimod_peptide_modifications_id(m.group(2))
                if i >= 0:
                    pos = 0
                    try:
                        pos = int(m.group(1))-1
                    except ValueError:
                        if 'C' == m.group(1):
                            pos = len(peptide_aminoacid_sequence) - 1
                    neutrons = element_count_and_mass_utils.unimod_peptide_modifications_extra_neutrons(i)
                    if neutrons > 0:
                        psm_neutrons_in_ms2_spectrum += \
                            element_count_and_mass_utils.unimod_peptide_modifications_extra_neutrons(i)
                    else:
                        peptide_modifications.append(i)
                        peptide_modification_positions.append(pos)
                else:
                    peptide_mass_undefined = True
                    self.unknown_modifications.add(modification_definition)

            peptide_element_counts = np.zeros(5, dtype=np.int16)  # C, N, O, H, S
            peptide_mass = 0
            try:
                peptide_mass = element_count_and_mass_utils.compute_peptide_mass(peptide_aminoacid_sequence,
                                                                                 peptide_modifications,
                                                                                 peptide_element_counts)
                psm_neutrons_in_ms2_spectrum /= peptide_element_counts[isotopic_pattern_utils.ELEMENT_ROW_INDEX]
            except KeyError:
                peptide_element_counts = np.zeros(5, dtype=np.int16)
                peptide_mass_undefined = True
                psm_neutrons_in_ms2_spectrum = 0
            except ValueError:
                print(peptide_aminoacid_sequence)
                exit(1)

            return {'experiment': os.path.basename(self.experiment_file_name),
                    'ms_run': ms_run_name,
                    'bins': ' '.join(peptide_bin_names),
                    'proteins': ' '.join(peptide_protein_names),
                    'peptide': f'{peptide_aminoacid_sequence} '
                               f'{[element_count_and_mass_utils.unimod_list()[m] for m in peptide_modifications]} '
                               f'{peptide_modification_positions}',
                    'psm_id': psm_spectrum_id,
                    'psm_mz': psm_mz,
                    'psm_charge': psm_charge,
                    'psm_neutrons': psm_neutrons_in_ms2_spectrum,
                    'psm_rank': 0,
                    'psm_precursor_id': 0,  # precursor_ms1_spectrum_id
                    'psm_precursor_mz': 0,  # precursor_ms1_spectrum_mz
                    'C': peptide_element_counts[0],
                    'N': peptide_element_counts[1],
                    'O': peptide_element_counts[2],
                    'H': peptide_element_counts[3],
                    'S': peptide_element_counts[4],
                    'peptide_mass': peptide_mass,
                    'flag_peptide_contains_sulfur': element_count_and_mass_utils.contains_sulfur(peptide_aminoacid_sequence),
                    'flag_peptide_has_modifications': len(peptide_modifications) > 0,
                    'flag_peptide_assigned_to_multiple_bins': len(peptide_bin_names) > 1,
                    'flag_peptide_assigned_to_multiple_proteins': len(peptide_protein_names) > 1,
                    'flag_peptide_mass_and_elements_undefined': peptide_mass_undefined,
                    'flag_psm_has_low_confidence': not psm_is_above_threshold,
                    'flag_psm_is_ambiguous': psm_is_ambiguous}

        # end of file reached
        # log problems
        if len(self.unknown_modifications):
            print(f'Warning: Encountered {len(self.unknown_modifications)} unknown modifications, '
                  f'such as {self.unknown_modifications}')
        raise StopIteration



UNIMOD='''C0H0N0O0S0	1_neutr_nterm	1
C0H0N0O0S0	2_neutr_nterm	2
C0H0N0O0S0	3_neutr_nterm	3
C0H0N0O0S0	4_neutr_cterm	4
C0H0N0O0S0	5_neutr_cterm	5
C0H0N0O0S0	6_neutr_cterm	6
C0H0N0O0S0	1x13CN	1
C0H0N0O0S0	2x13C_NTerm	2
C0H0N0O0S0	3X13C_NTerm	3
C0H0N0O0S0	4x13C_NTerm	4
C0H0N0O0S0	5x13C_NTerm	5
C0H0N0O0S0	6x13C_NTerm	6
C0H0N0O0S0	2x13C_CTerm	2
C0H0N0O0S0	3x13C_CTerm	3
C0H0N0O0S0	4x13C_CTerm	4
C0H0N0O0S0	5x13C_CTerm	5
C0H0N0O0S0	6x13C_CTerm	6
C2H2N0O1S0	Acetyl
C0H1N1O-1S0	Amidated
C10H14N2O2S1	Biotin
C2H3N1O1S0	Carbamidomethyl
C1H1N1O1S0	Carbamyl
C2H2N0O2S0	Carboxymethyl
C0H-1N-1O1S0	Deamidated
C22H38N4O6S1	ICAT-G
C-1H-2N0O1S-1	Met->Hse
C-1H-4N0O0S-1	Met->Hsl
C20H34N4O5S1	ICAT-D
C5H9N1O1S0	NIPCAM
C18H30N4O5S1	PEO-Iodoacetyl-LC-Biotin
C0H-2N0O-1S0	Dehydrated
C3H5N1O1S0	Propionamide
C7H5N1O1S0	Pyridylacetyl
C2H0N0O1S0	Pyro-carbamidomethyl
C0H-2N0O-1S0	Glu->pyro-Glu
C0H-3N-1O0S0	Gln->pyro-Glu
C6H9N1O2S0	SMA
C7H7N1O0S0	Pyridylethyl
C1H2N0O0S0	Methyl
C0H0N0O1S0	Oxidation
C2H4N0O0S0	Dimethyl
C3H6N0O0S0	Trimethyl
C1H2N0O0S1	Methylthio
C0H0N0O3S1	Sulfo
C6H10N0O5S0	Hex
C8H12N0O1S2	Lipoyl
C8H13N1O5S0	HexNAc
C15H24N0O0S0	Farnesyl
C14H26N0O1S0	Myristoyl
C16H30N0O1S0	Palmitoyl
C20H32N0O0S0	GeranylGeranyl
C51H96N0O5S0	Tripalmitate
C1H2N2O0S0	Guanidinyl
C9H16N0O2S0	HNE
C6H8N0O6S0	Glucuronyl
C10H15N3O6S1	Glutathione
C3H4N0O1S0	Propionyl
C13H31N3O1S0	GIST-Quat
C4H4N0O3S0	Succinyl
C10H18N4O2S1	probiotinhydrazide
C0H-2N0O1S0	Pro->pyro-Glu
C-2H-1N-1O1S0	His->Asn
C-2H-2N-2O2S0	His->Asp
C-1H0N0O2S0	Trp->Hydroxykynurenin
C3H4N0O0S0	Delta:H(4)C(3)
C2H4N0O0S0	Delta:H(4)C(2)
C0H-2N0O0S-1	Cys->Dha
C-1H-5N-3O1S0	Arg->GluSA
C0H0N0O3S0	Trioxidation
C10H15N3O1S1	Iminobiotin
C16H26N4O2S1	ESP
C16H25N3O3S1	NHS-LC-Biotin
C25H39N5O6S3	EDT-maleimide-PEO-biotin
C3H4N2O0S0	IMID
C10H15N3O2S1	Lysbiotinhydrazide
C0H-1N1O2S0	Nitro
C10H17N3O3S0	ICAT-C
C2H2N0O0S0	Delta:H(2)C(2)
C-1H0N0O1S0	Trp->Kynurenin
C0H-3N-1O1S0	Lys->Allysine
C6H9N1O2S1	FormylMet
C6H7N1O2S0	Nethylmaleimide
C16H26N4O3S1	OxLysBiotinRed
C16H24N4O3S1	OxLysBiotin
C16H29N5O3S1	OxProBiotinRed
C16H27N5O3S1	OxProBiotin
C15H22N2O3S1	OxArgBiotin
C15H24N2O3S1	OxArgBiotinRed
C20H34N4O4S3	EDT-iodoacetyl-PEO-biotin
C4H6N2O2S0	GG
C1H0N0O1S0	Formyl
C3H4N0O1S1	Xlink:DTSSP[88]
C10H10N0O6S0	Xlink:EGS[226]
C22H13N1O6S0	Fluorescein
C14H24N0O1S0	Myristoleyl
C-1H-2N0O-1S0	Pro->Pyrrolidinone
C14H22N0O1S0	Myristoyl+Delta:H(-4)
C7H4N0O1S0	Benzoyl
C46H76N2O35S0	Hex(5)HexNAc(2)
C12H11N1O2S1	Dansyl
C-1H-2N0O-2S0	a-type-ion
C2H3N1O0S0	Amidine
C14H23N1O9S0	HexNAc(1)dHex(1)
C16H26N2O10S0	HexNAc(2)
C18H30N0O15S0	Hex(3)
C20H33N1O13S0	HexNAc(1)dHex(2)
C20H33N1O14S0	Hex(1)HexNAc(1)dHex(1)
C22H36N2O14S0	HexNAc(2)dHex(1)
C22H36N2O15S0	Hex(1)HexNAc(2)
C25H40N2O18S0	Hex(1)HexNAc(1)NeuAc(1)
C28H46N2O18S0	HexNAc(2)dHex(2)
C27H44N2O19S0	Hex(1)HexNAc(2)Pent(1)
C28H46N2O19S0	Hex(1)HexNAc(2)dHex(1)
C28H46N2O20S0	Hex(2)HexNAc(2)
C31H51N1O24S0	Hex(3)HexNAc(1)Pent(1)
C33H54N2O23S0	Hex(1)HexNAc(2)dHex(1)Pent(1)
C34H56N2O23S0	Hex(1)HexNAc(2)dHex(2)
C33H54N2O24S0	Hex(2)HexNAc(2)Pent(1)
C34H56N2O24S0	Hex(2)HexNAc(2)dHex(1)
C34H56N2O25S0	Hex(3)HexNAc(2)
C36H57N3O26S0	Hex(1)HexNAc(1)NeuAc(2)
C6H3N1O2S1	NBS
C9H19N2O1S0	QAT
C15H22N0O1S0	BHT
C2H4N0O-1S1	Delta:H(4)C(2)O(-1)S(1)
C4H9N1O-1S1	DAET
C-1H0N0O-1S0	Pro->Pyrrolidone
C8H4N0O2S0	HPG
C16H10N0O5S0	2HPG
C10H6N2O1S0	AccQTag
C10H20N2O1S0	EQAT
C2H4N0O-1S2	Ethanedithiol
C6H6N0O1S0	Delta:H(6)C(6)O(1)
C3H4N0O1S0	Delta:H(4)C(3)O(1)
C3H2N0O0S0	Delta:H(2)C(3)
C6H4N0O0S0	Delta:H(4)C(6)
C6H8N0O2S0	Delta:H(8)C(6)O(2)
C4H7N1O1S0	NEIAA
C4H6N0O1S0	Crotonaldehyde
C0H1N1O0S0	Amino
C9H13N1O2S1	Argbiotinhydrazide
C7H5N1O3S2	SPITC
C19H22N0O7S0	Cytopiloyne
C19H24N0O8S0	Cytopiloyne+water
C7H7N1O-1S1	PET
C3H4N0O4S1	CAF
C5H4N0O2S0	Xlink:BS2G[96]
C0H-1N1O1S0	Nitrosyl
C8H9N1O2S1	AEBS
C2H4N0O1S0	Ethanolyl
C4H6N0O2S0	HMVK
C2H4N0O0S0	Ethyl
C1H1N-1O1S0	Methyl+Deamidated
C2H5N0O0S0	Delta:H(5)C(2)
C6H5N1O2S1	SulfanilicAcid
C16H28N4O3S1	Biotin-PEO-Amine
C0H-2N0O1S0	Trp->Oxolactone
C19H32N4O3S2	Biotin-HPDP
C5H7N1O2S1	CAMthiopropanoyl
C14H22N4O3S1	IED-Biotin
C6H10N0O4S0	dHex
C1H0N0O2S0	Carboxy
C10H10N2O2S0	Bromobimane
C11H6N0O2S0	Menadione
C2H4N0O1S1	DeStreak
C56H92N4O39S0	dHex(1)Hex(3)HexNAc(4)
C62H102N4O44S0	dHex(1)Hex(4)HexNAc(4)
C68H112N4O49S0	dHex(1)Hex(5)HexNAc(4)
C50H82N4O35S0	Hex(3)HexNAc(4)
C56H92N4O40S0	Hex(4)HexNAc(4)
C62H102N4O45S0	Hex(5)HexNAc(4)
C3H5N1O2S1	Cysteinyl
C-6H-12N-2O-1S0	Lys-loss
C5H5N1O2S0	Nmethylmaleimide
C37H44N4O6S1	CyDye-Cy3
C6H6N0O0S0	DimethylpyrroleAdduct
C5H2N0O0S0	Delta:H(2)C(5)
C3H2N0O1S0	Delta:H(2)C(3)O(1)
C6H9N1O3S0	Nethylmaleimide+water
C3H5N1O0S1	Xlink:DTBP[87]
C9H18N0O2S0	HNE+Delta:H(2)
C10H16N4O1S1	Thrbiotinhydrazide
C1H3N1O-1S0	Methylamine
C6H11N1O3S2	CarbamidomethylDTT
C6H3N1O1S0	ICPL
C-1H-2N-2O0S0	Arg->Orn
C0H-1N0O0S0	Dehydro
C7H14N2O1S0	Diphthamide
C15H24N0O1S0	Hydroxyfarnesyl
C37H68N0O4S0	Diacylglycerol
C3H4N0O2S0	Carboxyethyl
C4H9N1O1S0	Hypusine
C20H26N0O0S0	Retinylidene
C0H-3N-1O2S0	Lys->AminoadipicAcid
C0H-3N-1O1S-1	Cys->PyruvicAcid
C0H-3N-1O0S0	Ammonia-loss
C33H38N4O6S0	Phycocyanobilin
C33H40N4O6S0	Phycoerythrobilin
C33H36N4O6S0	Phytochromobilin
C0H-2N0O2S0	Quinone
C12H20N0O11S0	Glucosylgalactosyl
C-6H-6N0O-1S0	Tyr->Dha
C0H-2N0O0S0	Didehydro
C0H-2N0O1S-1	Cys->Oxoalanine
C0H-1N-1O0S0	Ser->LacticAcid
C10H14N2O6S0	GluGlu
C5H7N1O3S0	Glu
C9H6N0O2S0	Hydroxycinnamyl
C5H8N0O5S0	Glycosyl
C43H86N0O2S0	Archaeol
C7H5N1O1S0	Phenylisocyanate
C1H2N0O1S0	Hydroxymethyl
C20H22N2O8S0	Dipyrrolylmethanemethyl
C0H0N0O-1S1	Carboxy->Thiocarboxy
C0H0N0O0S1	Sulfide
C3H2N0O2S0	PyruvicAcidIminyl
C0H0N0O2S0	Dioxidation
C8H14N0O1S0	Octanoyl
C16H28N0O1S0	Palmitoleyl
C27H44N0O0S0	Cholesterol
C20H24N0O0S0	Didehydroretinylidene
C17H26N0O4S0	CHDH
C6H7N1O1S0	Methylpyrroline
C1H-1N1O0S0	Cyano
C1H2N2O0S0	Amidino
C3H7N0O1S0	Hydroxytrimethyl
C0H0N0O-1S0	Deoxy
C36H37N3O20S0	Microcin
C10H18N0O1S0	Decanoyl
C15H21N3O9S0	GluGluGlu
C20H28N4O12S0	GluGluGluGlu
C6H11N1O4S0	HexN
C8H14N2O1S0	Xlink:DMP[154]
C13H5N1O0S0	NDA
C13H32N3O1S0	TMAB
C21H15N3O5S1	FTC
C2H5N1O-1S1	AEC-MAEC
C21H24N0O4S0	BADGE
C28H48N0O24S0	Hep
C38H44N4O6S1	CyDye-Cy5
C8H8N1O0S0	DHP
C15H22N0O2S0	BHTOH
C5H7N1O3S0	Nmethylmaleimide+water
C7H6N2O1S0	PyMIC
C20H28N0O4S0	LG-lactam-K
C32H22N2O6S2	BisANS
C5H8N0O0S0	Piperidine
C4H8N0O0S0	Diethyl
C20H28N0O5S0	LG-Hlactam-K
C14H29N1O1S0	C8-QAT
C12H20N0O10S0	Hex(2)
C19H26N-2O4S0	LG-lactam-R
C28H38N0O6S0	Withaferin
C19H26N-2O5S0	LG-Hlactam-R
C23H35N5O7S1	Maleimide-PEO2-Biotin
C22H36N4O4S1	Sulfo-NHS-LC-LC-Biotin
C24H13N1O7S0	FNEM
C9H14N1O4S1	PropylNAGthiazoline
C-1H-4N0O0S-1	Dethiomethyl
C16H29N7O4S0	LRGG
C35H54N4O4S1	Biotin:Cayman-10141
C36H60N4O5S1	Biotin:Cayman-10013
C0H0N0O1S0	Ala->Ser
C1H2N0O1S0	Ala->Thr
C1H0N0O2S0	Ala->Asp
C2H2N0O0S0	Ala->Pro
C-1H-2N0O0S0	Ala->Gly
C2H2N0O2S0	Ala->Glu
C2H4N0O0S0	Ala->Val
C6H4N0O0S-1	Cys->Phe
C0H0N0O1S-1	Cys->Ser
C8H5N1O0S-1	Cys->Trp
C6H4N0O1S-1	Cys->Tyr
C3H7N3O0S-1	Cys->Arg
C-1H-2N0O0S-1	Cys->Gly
C-1H0N0O-2S0	Asp->Ala
C2H2N2O-2S0	Asp->His
C0H1N1O-1S0	Asp->Asn
C-2H-2N0O-2S0	Asp->Gly
C5H4N0O-1S0	Asp->Tyr
C1H2N0O0S0	Asp->Glu
C1H4N0O-2S0	Asp->Val
C-2H-2N0O-2S0	Glu->Ala
C0H1N1O-1S0	Glu->Gln
C-1H-2N0O0S0	Glu->Asp
C1H5N1O-2S0	Glu->Lys
C-3H-4N0O-2S0	Glu->Gly
C0H2N0O-2S0	Glu->Val
C-6H-4N0O1S0	Phe->Ser
C-6H-4N0O0S1	Phe->Cys
C-3H2N0O0S0	Phe->Xle
C0H0N0O1S0	Phe->Tyr
C-4H0N0O0S0	Phe->Val
C1H2N0O0S0	Gly->Ala
C1H2N0O1S0	Gly->Ser
C9H7N1O0S0	Gly->Trp
C3H4N0O2S0	Gly->Glu
C3H6N0O0S0	Gly->Val
C2H2N0O2S0	Gly->Asp
C1H2N0O0S1	Gly->Cys
C4H9N3O0S0	Gly->Arg
C-1H0N-2O0S0	His->Pro
C3H2N-2O1S0	His->Tyr
C-1H1N-1O1S0	His->Gln
C6H3N1O1S0	NIC
C0H5N1O0S0	His->Arg
C0H4N-2O0S0	His->Xle
C-3H-6N0O0S0	Xle->Ala
C-2H-4N0O1S0	Xle->Thr
C-2H-5N1O1S0	Xle->Asn
C0H1N1O0S0	Xle->Lys
C-2H-5N-1O1S0	Lys->Thr
C-2H-6N0O1S0	Lys->Asn
C-1H-5N-1O2S0	Lys->Glu
C-1H-4N0O1S0	Lys->Gln
C-1H-3N-1O0S1	Lys->Met
C0H0N2O0S0	Lys->Arg
C0H-1N-1O0S0	Lys->Xle
C-3H-6N0O1S0	Xle->Ser
C3H-2N0O0S0	Xle->Phe
C5H-1N1O0S0	Xle->Trp
C-1H-4N0O0S0	Xle->Pro
C-1H-2N0O0S0	Xle->Val
C0H-4N2O0S0	Xle->His
C-1H-3N1O1S0	Xle->Gln
C-1H-2N0O0S1	Xle->Met
C0H1N3O0S0	Xle->Arg
C-1H-2N0O1S-1	Met->Thr
C1H3N3O0S-1	Met->Arg
C1H3N1O0S-1	Met->Lys
C1H2N0O0S-1	Met->Xle
C0H0N0O0S-1	Met->Val
C-1H-1N-1O0S0	Asn->Ser
C0H1N-1O0S0	Asn->Thr
C2H6N0O-1S0	Asn->Lys
C5H3N-1O0S0	Asn->Tyr
C2H1N1O-1S0	Asn->His
C0H-1N-1O1S0	Asn->Asp
C2H5N-1O-1S0	Asn->Xle
C-2H-2N0O1S0	Pro->Ser
C-2H-2N0O0S0	Pro->Ala
C1H0N2O0S0	Pro->His
C0H1N1O1S0	Pro->Gln
C-1H0N0O1S0	Pro->Thr
C1H5N3O0S0	Pro->Arg
C1H4N0O0S0	Pro->Xle
C0H-1N-1O-1S0	Gln->Pro
C1H4N0O-1S0	Gln->Lys
C0H-1N-1O1S0	Gln->Glu
C1H-1N1O-1S0	Gln->His
C1H4N2O-1S0	Gln->Arg
C1H3N-1O-1S0	Gln->Xle
C-3H-7N-3O1S0	Arg->Ser
C5H-2N-2O0S0	Arg->Trp
C-2H-5N-3O1S0	Arg->Thr
C-1H-5N-3O0S0	Arg->Pro
C0H0N-2O0S0	Arg->Lys
C0H-5N-1O0S0	Arg->His
C-1H-4N-2O1S0	Arg->Gln
C-1H-3N-3O0S1	Arg->Met
C-3H-7N-3O0S1	Arg->Cys
C0H-1N-3O0S0	Arg->Xle
C-4H-9N-3O0S0	Arg->Gly
C6H4N0O-1S0	Ser->Phe
C0H0N0O-1S0	Ser->Ala
C8H5N1O-1S0	Ser->Trp
C1H2N0O0S0	Ser->Thr
C1H1N1O0S0	Ser->Asn
C2H2N0O-1S0	Ser->Pro
C6H4N0O0S0	Ser->Tyr
C0H0N0O-1S1	Ser->Cys
C3H7N3O-1S0	Ser->Arg
C3H6N0O-1S0	Ser->Xle
C-1H-2N0O-1S0	Ser->Gly
C-1H-2N0O0S0	Thr->Ser
C-1H-2N0O-1S0	Thr->Ala
C0H-1N1O0S0	Thr->Asn
C2H5N1O-1S0	Thr->Lys
C1H0N0O-1S0	Thr->Pro
C1H2N0O-1S1	Thr->Met
C2H4N0O-1S0	Thr->Xle
C2H5N3O-1S0	Thr->Arg
C4H0N0O0S0	Val->Phe
C-2H-4N0O0S0	Val->Ala
C0H-2N0O2S0	Val->Glu
C0H0N0O0S1	Val->Met
C-1H-4N0O2S0	Val->Asp
C1H2N0O0S0	Val->Xle
C-3H-6N0O0S0	Val->Gly
C-8H-5N-1O1S0	Trp->Ser
C-8H-5N-1O0S1	Trp->Cys
C-5H2N2O0S0	Trp->Arg
C-9H-7N-1O0S0	Trp->Gly
C-5H1N-1O0S0	Trp->Xle
C0H0N0O-1S0	Tyr->Phe
C-6H-4N0O0S0	Tyr->Ser
C-5H-3N1O0S0	Tyr->Asn
C-3H-2N2O-1S0	Tyr->His
C-5H-4N0O1S0	Tyr->Asp
C-6H-4N0O-1S1	Tyr->Cys
C18H31N1O4S0	NA-LNO2
C18H33N1O4S0	NA-OA-NO2
C6H10N0O4S2	CarboxymethylDTT
C9H14N0O1S0	HNE-Delta:H(2)O
C9H14N0O2S0	4-ONE
C4H8N0O1S2	BEMAD_ST
C2H5N1O0S0	Ethanolamine
C4H8N0O2S1	BEMAD_C
C12H20N2O2S0	TMT
C10H10N3O3S1	NO_SMX_SEMD
C9H12N0O1S0	4-ONE+Delta:H(-2)O(-1)
C10H10N3O4S1	NO_SMX_SMCT
C10H9N3O4S1	NO_SMX_SIMD
C3H2N0O3S0	Malonyl
C7H4N0O4S1	3sulfo
C6H1N3O6S0	TNBS
C29H38N8O6S1	Biotin-phenacyl
C16H16N0O2S0	lapachenole
C4H3N1O2S0	maleimide
C-5H-9N-1O-1S-1	Met-loss
C-3H-7N-1O0S-1	Met-loss+Acetyl
C11H8N0O2S0	Menadione-HQ
C-2H-2N0O-2S0	Gly-loss+Amide
C3H-1N1O2S0	Arg->Npo
C5H11N1O0S0	Pentylamine
C15H25N3O2S1	Biotin:Thermo-21345
C3H4N0O2S0	Dihydroxyimidazolidine
C6H0N2O4S0	Xlink:DFDNB
C37H38N4O7S1	Cy3b-maleimide
C14H23N1O10S0	Hex(1)HexNAc(1)
C10H8N2O4S0	Xlink:BMOE
C21H37N5O6S1	Biotin:Thermo-21360
C20H32N6O8S0	EQIGG
C5H4N5O1S0	cGMP+RMP-loss
C7H12N2O1S0	mTRAQ
C16H10N0O4S0	Arg2PG
C10H23N3O0S0	spermine
C3H2N0O1S0	MG-H1
C2H0N0O1S0	G-H1
C39H48N4O15S4	DyLight-maleimide
C32H58N2O15S0	Methyl-PEO12-Maleimide
C20H21N3O5S0	MDCC
C23H37N9O10S0	QQQTGG
C23H36N8O11S0	QEQTGG
C6H4N0O2S0	HydroxymethylOP
C34H45N7O7S1	Biotin:Thermo-21325
C26H42N8O7S0	Biotin-PEG-PRA
C-1H-3N3O0S-1	Met->Aha
C1H-2N0O0S-1	Met->Hpg
C17H24N0O9S0	4AcAllylGal
C-1H-4N0O1S1	Lys->CamCys
C-4H-1N1O1S1	Phe->CamCys
C-1H-2N0O1S1	Leu->MetOx
C-1H-3N-1O1S1	Lys->MetOx
C6H10N0O6S0	Galactosyl
C17H27N3O3S0	Xlink:SMCC[321]
C10H16N2O4S0	Bacillosamine
C9H14N1O1S1	MTSL
C25H45N5O4S1	HNE-BAHAH
C3H4N0O2S0	Ethoxyformyl
C4H4N0O3S0	Methylmalonylation
C35H52N10O9S2	AROD
C1H3N1O0S-1	Cys->methylaminoAla
C2H5N1O0S-1	Cys->ethylaminoAla
C2H5N1O-1S0	ethylamino
C2H4N0O0S1	MercaptoEthanol
C27H32N5O3S0	Atto495Maleimide
C19H30N6O10S0	AMTzHexNAc2
C2H3N-1O1S0	Ethyl+Deamidated
C37H55N11O12S0	VFQQQTGG
C53H81N13O19S0	VIEVYQEQTGG
C6H3N2O4S1	DNPS
C22H26N4O5S1	SulfoGMBS
C13H21N3O3S0	DimethylamineGMBS
C20H26N0O3S0	LG-anhydrolactam
C20H28N0O3S0	LG-pyrrole
C20H26N0O2S0	LG-anhyropyrrole
C6H8N0O4S0	3-deoxyglucosone
C4H4N0O4S0	2-succinyl
C3H3N1O-1S0	Propargylamine
C90H137N21O37S1	SUMO2135
C150H224N38O60S1	SUMO3549
C10H9N1O1S0	serotonylation
C8H7N1O0S1	BITC
C2H4N1O1S0	Carbofuran
C9H9N1O0S1	PEITC
C6H9N1O2S1	thioacylPA
C37H59N7O23S0	maleimide3
C49H79N7O33S0	maleimide5
C22H27N7O4S0	Puromycin
C6H8N0O5S0	glucosone
C14H25N3O2S1	cysTMT
C0H-1N-1O0S0	ISD_z+2_ion
C0H3N1O0S0	Ammonium
C20H27N5O5S1	Biotin:Sigma-B1267
C0H1N1O3S1	sulfo+amino
C4H5N5O1S-1	AHA-Alkyne
C26H37N11O14S-1	AHA-Alkyne-KDDDD
C22H16N0O11S0	EGCG1
C15H11N0O6S0	EGCG2
C1H0N0O0S0	Thiazolidine
C89H122N18O31S1	DEDGFLYMVYASQETFG
C23H33N5O7S1	Biotin:Invitrogen-M1602
C8H12N0O3S0	Xlink:DSS[156]
C6H9N1O1S0	DMPO
C3H5N1O2S0	glycidamide
C16H27N3O3S0	Ahx2+Hsl
C8H10N0O2S0	ICDID
C10H12N0O7S0	Xlink:EGS[244]
C4H4N0O5S0	Xlink:DST[132]
C6H8N0O3S2	Xlink:DTSSP[192]
C12H15N1O4S0	Xlink:SMCC[237]
C7H5N1O2S0	2-nitrobenzyl
C7H12N2O1S0	Xlink:DMP[140]
C4H5N1O3S0	Xlink:EGS[115]
C9H6N2O0S1	Thiadiazole
C10H16N2O2S0	Biotin:Thermo-88310
C23H37N5O8S1	Biotin:Thermo-21901+H2O
C4H9N1O0S0	Deoxyhypusine
C6H11N1O0S0	Acetyldeoxyhypusine
C6H11N1O1S0	Acetylhypusine
C0H0N0O0S1	Ala->Cys
C6H4N0O0S0	Ala->Phe
C3H2N2O0S0	Ala->His
C3H6N0O0S0	Ala->Xle
C3H7N1O0S0	Ala->Lys
C2H4N0O0S1	Ala->Met
C1H1N1O1S0	Ala->Asn
C2H3N1O1S0	Ala->Gln
C3H7N3O0S0	Ala->Arg
C8H5N1O0S0	Ala->Trp
C6H4N0O1S0	Ala->Tyr
C0H0N0O0S-1	Cys->Ala
C1H0N0O2S-1	Cys->Asp
C2H2N0O2S-1	Cys->Glu
C3H2N2O0S-1	Cys->His
C3H6N0O0S-1	Cys->Xle
C3H7N1O0S-1	Cys->Lys
C2H4N0O0S0	Cys->Met
C1H1N1O1S-1	Cys->Asn
C2H2N0O0S-1	Cys->Pro
C2H3N1O1S-1	Cys->Gln
C1H2N0O1S-1	Cys->Thr
C2H4N0O0S-1	Cys->Val
C-1H0N0O-2S1	Asp->Cys
C5H4N0O-2S0	Asp->Phe
C2H6N0O-2S0	Asp->Xle
C2H7N1O-2S0	Asp->Lys
C1H4N0O-2S1	Asp->Met
C1H2N0O-2S0	Asp->Pro
C1H3N1O-1S0	Asp->Gln
C2H7N3O-2S0	Asp->Arg
C-1H0N0O-1S0	Asp->Ser
C0H2N0O-1S0	Asp->Thr
C7H5N1O-2S0	Asp->Trp
C-2H-2N0O-2S1	Glu->Cys
C4H2N0O-2S0	Glu->Phe
C1H0N2O-2S0	Glu->His
C1H4N0O-2S0	Glu->Xle
C0H2N0O-2S1	Glu->Met
C-1H-1N1O-1S0	Glu->Asn
C0H0N0O-2S0	Glu->Pro
C1H5N3O-2S0	Glu->Arg
C-2H-2N0O-1S0	Glu->Ser
C-1H0N0O-1S0	Glu->Thr
C6H3N1O-2S0	Glu->Trp
C4H2N0O-1S0	Glu->Tyr
C-6H-4N0O0S0	Phe->Ala
C-5H-4N0O2S0	Phe->Asp
C-4H-2N0O2S0	Phe->Glu
C-7H-6N0O0S0	Phe->Gly
C-3H-2N2O0S0	Phe->His
C-3H3N1O0S0	Phe->Lys
C-4H0N0O0S1	Phe->Met
C-5H-3N1O1S0	Phe->Asn
C-4H-2N0O0S0	Phe->Pro
C-4H-1N1O1S0	Phe->Gln
C-3H3N3O0S0	Phe->Arg
C-5H-2N0O1S0	Phe->Thr
C2H1N1O0S0	Phe->Trp
C7H6N0O0S0	Gly->Phe
C4H4N2O0S0	Gly->His
C4H8N0O0S0	Gly->Xle
C4H9N1O0S0	Gly->Lys
C3H6N0O0S1	Gly->Met
C2H3N1O1S0	Gly->Asn
C3H4N0O0S0	Gly->Pro
C3H5N1O1S0	Gly->Gln
C2H4N0O1S0	Gly->Thr
C7H6N0O1S0	Gly->Tyr
C-3H-2N-2O0S0	His->Ala
C-3H-2N-2O0S1	His->Cys
C-1H0N-2O2S0	His->Glu
C3H2N-2O0S0	His->Phe
C-4H-4N-2O0S0	His->Gly
C0H5N-1O0S0	His->Lys
C-1H2N-2O0S1	His->Met
C-3H-2N-2O1S0	His->Ser
C-2H0N-2O1S0	His->Thr
C-1H2N-2O0S0	His->Val
C5H3N-1O0S0	His->Trp
C-3H-6N0O0S1	Xle->Cys
C-2H-6N0O2S0	Xle->Asp
C-1H-4N0O2S0	Xle->Glu
C-4H-8N0O0S0	Xle->Gly
C3H-2N0O1S0	Xle->Tyr
C-3H-7N-1O0S0	Lys->Ala
C-3H-7N-1O0S1	Lys->Cys
C-2H-7N-1O2S0	Lys->Asp
C3H-3N-1O0S0	Lys->Phe
C-4H-9N-1O0S0	Lys->Gly
C0H-5N1O0S0	Lys->His
C-1H-5N-1O0S0	Lys->Pro
C-3H-7N-1O1S0	Lys->Ser
C-1H-3N-1O0S0	Lys->Val
C5H-2N0O0S0	Lys->Trp
C3H-3N-1O1S0	Lys->Tyr
C-2H-4N0O0S-1	Met->Ala
C-2H-4N0O0S0	Met->Cys
C-1H-4N0O2S-1	Met->Asp
C0H-2N0O2S-1	Met->Glu
C4H0N0O0S-1	Met->Phe
C-3H-6N0O0S-1	Met->Gly
C1H-2N2O0S-1	Met->His
C-1H-3N1O1S-1	Met->Asn
C0H-2N0O0S-1	Met->Pro
C0H-1N1O1S-1	Met->Gln
C-2H-4N0O1S-1	Met->Ser
C6H1N1O0S-1	Met->Trp
C4H0N0O1S-1	Met->Tyr
C-1H-1N-1O-1S0	Asn->Ala
C-1H-1N-1O-1S1	Asn->Cys
C1H1N-1O1S0	Asn->Glu
C5H3N-1O-1S0	Asn->Phe
C-2H-3N-1O-1S0	Asn->Gly
C1H3N-1O-1S1	Asn->Met
C1H1N-1O-1S0	Asn->Pro
C1H2N0O0S0	Asn->Gln
C2H6N2O-1S0	Asn->Arg
C1H3N-1O-1S0	Asn->Val
C7H4N0O-1S0	Asn->Trp
C-2H-2N0O0S1	Pro->Cys
C-1H-2N0O2S0	Pro->Asp
C0H0N0O2S0	Pro->Glu
C4H2N0O0S0	Pro->Phe
C-3H-4N0O0S0	Pro->Gly
C1H5N1O0S0	Pro->Lys
C0H2N0O0S1	Pro->Met
C-1H-1N1O1S0	Pro->Asn
C0H2N0O0S0	Pro->Val
C6H3N1O0S0	Pro->Trp
C4H2N0O1S0	Pro->Tyr
C-2H-3N-1O-1S0	Gln->Ala
C-2H-3N-1O-1S1	Gln->Cys
C-1H-3N-1O1S0	Gln->Asp
C4H1N-1O-1S0	Gln->Phe
C-3H-5N-1O-1S0	Gln->Gly
C0H1N-1O-1S1	Gln->Met
C-1H-2N0O0S0	Gln->Asn
C-2H-3N-1O0S0	Gln->Ser
C-1H-1N-1O0S0	Gln->Thr
C0H1N-1O-1S0	Gln->Val
C6H2N0O-1S0	Gln->Trp
C4H1N-1O0S0	Gln->Tyr
C-3H-7N-3O0S0	Arg->Ala
C-2H-7N-3O2S0	Arg->Asp
C-1H-5N-3O2S0	Arg->Glu
C-2H-6N-2O1S0	Arg->Asn
C-1H-3N-3O0S0	Arg->Val
C3H-3N-3O1S0	Arg->Tyr
C3H-3N-3O0S0	Arg->Phe
C1H0N0O1S0	Ser->Asp
C2H2N0O1S0	Ser->Glu
C3H2N2O-1S0	Ser->His
C3H7N1O-1S0	Ser->Lys
C2H4N0O-1S1	Ser->Met
C2H3N1O0S0	Ser->Gln
C2H4N0O-1S0	Ser->Val
C-1H-2N0O-1S1	Thr->Cys
C0H-2N0O1S0	Thr->Asp
C1H0N0O1S0	Thr->Glu
C5H2N0O-1S0	Thr->Phe
C-2H-4N0O-1S0	Thr->Gly
C2H0N2O-1S0	Thr->His
C1H1N1O0S0	Thr->Gln
C1H2N0O-1S0	Thr->Val
C7H3N1O-1S0	Thr->Trp
C5H2N0O0S0	Thr->Tyr
C-2H-4N0O0S1	Val->Cys
C1H-2N2O0S0	Val->His
C1H3N1O0S0	Val->Lys
C-1H-3N1O1S0	Val->Asn
C0H-2N0O0S0	Val->Pro
C0H-1N1O1S0	Val->Gln
C1H3N3O0S0	Val->Arg
C-2H-4N0O1S0	Val->Ser
C-1H-2N0O1S0	Val->Thr
C6H1N1O0S0	Val->Trp
C4H0N0O1S0	Val->Tyr
C-8H-5N-1O0S0	Trp->Ala
C-7H-5N-1O2S0	Trp->Asp
C-6H-3N-1O2S0	Trp->Glu
C-2H-1N-1O0S0	Trp->Phe
C-5H-3N1O0S0	Trp->His
C-5H2N0O0S0	Trp->Lys
C-6H-1N-1O0S1	Trp->Met
C-7H-4N0O1S0	Trp->Asn
C-6H-3N-1O0S0	Trp->Pro
C-6H-2N0O1S0	Trp->Gln
C-7H-3N-1O1S0	Trp->Thr
C-6H-1N-1O0S0	Trp->Val
C-2H-1N-1O1S0	Trp->Tyr
C-6H-4N0O-1S0	Tyr->Ala
C-4H-2N0O1S0	Tyr->Glu
C-7H-6N0O-1S0	Tyr->Gly
C-3H3N1O-1S0	Tyr->Lys
C-4H0N0O-1S1	Tyr->Met
C-4H-2N0O-1S0	Tyr->Pro
C-4H-1N1O0S0	Tyr->Gln
C-3H3N3O-1S0	Tyr->Arg
C-5H-2N0O0S0	Tyr->Thr
C-4H0N0O-1S0	Tyr->Val
C2H1N1O-1S0	Tyr->Trp
C-3H2N0O-1S0	Tyr->Xle
C7H9N5O2S0	AHA-SS
C9H12N6O3S0	AHA-SS_CAM
C25H36N6O4S2	Biotin:Thermo-33033
C25H34N6O4S2	Biotin:Thermo-33033-H
C5H6N0O4S0	2-monomethylsuccinyl
C7H6N0O1S0	Saligenin
C4H8N2O1S0	Ub-Br2
C7H13N2O3S0	Ub-VME
C9H14N3O2S0	Ub-amide
C31H29N6O7S0	Ub-fluorescein
C6H8N0O4S0	2-dimethylsuccinyl
C2H3N1O1S0	Gly
C9H13N3O5S0	pupylation
C4H7N1O2S1	HCysteinyl
C47H60N23O10S0	UgiJoullie
C4H7N1O1S1	HCysThiolactone
C7H10N2O2S0	UgiJoullieProGly
C13H11N3O1S0	Dipyridyl
C4H2N0O1S0	Furan
C8H4N0O2S0	Difuran
C18H17N1O1S0	BMP-piperidinol
C14H20N4O4S0	UgiJoullieProGlyProGly
C-6H-12N-4O-1S0	Arg-loss
C6H12N4O1S0	Arg
C25H40N2O18S1	IMEHex(2)NeuAc(1)
C4H6N0O1S0	Butyryl
C4H6N2O2S0	Dicarbamidomethyl
C9H14N4O4S0	GGQ
C13H21N5O6S0	QTGG
C6H12N2O1S0	Lys
C44H68N4O32S0	NeuAc
C11H17N1O9S0	NeuGc
C3H6N0O0S0	Propyl
C9H8N0O1S0	Propiophenone
C7H4N0O2S0	PS_Hapten
C37H45N4O9S2	Cy3-maleimide
C3H6N0O1S0	Delta:H(6)C(3)O(1)
C6H8N0O1S0	Delta:H(8)C(6)O(1)
C13H22N4O2S1	biotinAcrolein298
C18H19N1O1S0	MM-diphenylpentanone
C18H18N0O2S0	EHD-diphenylpentanone
C8H8N2O0S0	benzylguanidine
C9H10N2O1S0	CarboxymethylDMAP
C23H39N5O9S1	Biotin:Thermo-21901+2H2O
C9H15N3O3S0	Xlink:BuUrBu[213]
C32H25N2O0S0	bisANS-sulfonates
C6H2N2O4S0	DNCB_hapten
C6H7N1O2S1	NEMsulfur
C0H0N0O2S1	SulfurDioxide
C6H9N1O3S1	NEMsulfurWater
C6H13N1O2S0	HN3_mustard
C5H11N1O1S0	HN2_mustard
C4H4N0O1S0	Crotonyl
C0H-4N0O-1S0	azole
C41H71N5O16S1	Biotin:Thermo-21911
C16H28N4O3S0	iodoTMT
C36H60N0O27S0	dHex(1)Hex(1)
C3H6N0O2S1	methylsulfonylethyl
C54H90N0O42S0	dHex(1)Hex(2)
C24H40N0O19S0	dHex(1)Hex(3)
C30H50N0O24S0	dHex(1)Hex(4)
C36H60N0O29S0	dHex(1)Hex(5)
C42H70N0O34S0	dHex(1)Hex(6)
C4H8N0O2S1	ethylsulfonylethyl
C8H8N0O2S1	phenylsulfonylethyl
C-1H-2N0O3S0	Homocysteic_acid
C0H1N1O0S0	Hydroxamic_acid
C6H7N1O3S0	Oxidation+NEM
C27H21N1O7S0	NHS-fluorescein
C8H7N1O1S0	Iodoacetanilide
C13H20N2O6S2	Dap-DSP
C11H17N1O7S0	MurNAc
C72H107N17O31S0	EEEDVIEVYQEQTGG
C69H102N18O30S0	EDEDTIDVFQQQTGG
C84H136N6O61S0	Hex(5)HexNAc(4)NeuAc(2)
C73H119N5O53S0	Hex(5)HexNAc(4)NeuAc(1)
C79H129N5O57S0	dHex(1)Hex(5)HexNAc(4)NeuAc(1)
C90H146N6O65S0	dHex(1)Hex(5)HexNAc(4)NeuAc(2)
C8H13N1O8S1	s-GlcNAc
C7H16N2O0S0	spermidine
C21H35N3O7S1	Biotin:Thermo-21330
C16H26N0O13S0	Hex(1)Pent(2)
C5H8N0O4S0	Pentose
C11H18N0O9S0	Hex(1)Pent(1)
C12H18N0O11S0	Hex(1)HexA(1)
C42H69N3O39S3	Hex(1)HexNAc(1)Sulf(1)
C17H27N1O13S0	Hex(1)NeuAc(1)
C17H27N1O14S0	Hex(1)NeuGc(1)
C24H39N3O15S0	HexNAc(3)
C19H30N2O13S0	HexNAc(1)NeuAc(1)
C19H30N2O14S0	HexNAc(1)NeuGc(1)
C23H37N1O18S0	Hex(2)NeuAc(1)
C21H35N1O14S0	Hex(1)HexNAc(1)dHex(1)Me(1)
C22H37N1O14S0	Hex(1)HexNAc(1)dHex(1)Me(2)
C8H13N1O2S0	Xlink:DSS[155]
C20H33N1O15S0	Hex(2)HexNAc(1)
C20H31N1O16S0	Hex(1)HexA(1)HexNAc(1)
C21H35N1O15S0	Hex(2)HexNAc(1)Me(1)
C21H34N0O17S0	Hex(1)Pent(3)
C22H35N1O17S0	Hex(1)NeuAc(1)Pent(1)
C20H33N1O18S1	Hex(2)HexNAc(1)Sulf(1)
C24H40N0O18S0	dHex(2)Hex(2)
C24H38N0O20S0	dHex(1)Hex(2)HexA(1)
C22H36N2O18S1	Hex(1)HexNAc(2)Sulf(1)
C24H40N0O20S0	Hex(4)
C39H64N2O28S0	dHex(1)Hex(2)HexNAc(2)Pent(1)
C39H63N3O28S0	Hex(2)HexNAc(2)NeuAc(1)
C39H64N2O29S0	Hex(3)HexNAc(2)Pent(1)
C40H66N2O30S0	Hex(4)HexNAc(2)
C43H71N1O33S0	dHex(1)Hex(4)HexNAc(1)Pent(1)
C45H74N2O33S0	dHex(1)Hex(3)HexNAc(2)Pent(1)
C45H73N3O33S0	Hex(3)HexNAc(2)NeuAc(1)
C45H74N2O34S0	Hex(4)HexNAc(2)Pent(1)
C47H77N3O34S0	Hex(3)HexNAc(3)Pent(1)
C102H168N4O76S0	dHex(1)Hex(4)HexNAc(2)Pent(1)
C50H83N1O40S0	Hex(7)HexNAc(1)
C51H83N3O38S0	Hex(4)HexNAc(2)NeuAc(1)
C52H86N2O39S0	dHex(1)Hex(5)HexNAc(2)
C53H87N3O38S0	dHex(1)Hex(3)HexNAc(3)Pent(1)
C50H82N4O38S1	Hex(3)HexNAc(4)Sulf(1)
C52H86N2O40S0	Hex(6)HexNAc(2)
C53H87N3O39S0	Hex(4)HexNAc(3)Pent(1)
C54H89N3O39S0	dHex(1)Hex(4)HexNAc(3)
C54H89N3O40S0	Hex(5)HexNAc(3)
C55H90N4O39S0	Hex(3)HexNAc(4)Pent(1)
C108H178N6O84S2	dHex(1)Hex(4)HexNAc(3)Sulf(1)
C57H94N2O43S0	dHex(1)Hex(5)HexNAc(2)Pent(1)
C56H93N1O45S0	Hex(8)HexNAc(1)
C58H95N3O42S0	dHex(1)Hex(3)HexNAc(3)Pent(2)
C59H97N3O42S0	dHex(2)Hex(3)HexNAc(3)Pent(1)
C56H92N4O42S1	dHex(1)Hex(3)HexNAc(4)Sulf(1)
C58H96N2O44S0	dHex(1)Hex(6)HexNAc(2)
C59H97N3O43S0	dHex(1)Hex(4)HexNAc(3)Pent(1)
C56H92N4O43S1	Hex(4)HexNAc(4)Sulf(1)
C58H96N2O45S0	Hex(7)HexNAc(2)
C60H99N3O43S0	dHex(2)Hex(4)HexNAc(3)
C59H97N3O44S0	Hex(5)HexNAc(3)Pent(1)
C59H96N4O44S0	Hex(4)HexNAc(3)NeuGc(1)
C60H99N3O44S0	dHex(1)Hex(5)HexNAc(3)
C61H100N4O43S0	dHex(1)Hex(3)HexNAc(4)Pent(1)
C58H95N5O43S1	Hex(3)HexNAc(5)Sulf(1)
C60H99N3O45S0	Hex(6)HexNAc(3)
C61H99N5O43S0	Hex(3)HexNAc(4)NeuAc(1)
C61H100N4O44S0	Hex(4)HexNAc(4)Pent(1)
C126H208N8O88S0	Hex(4)HexNAc(4)Me(2)Pent(1)
C63H103N3O46S0	dHex(1)Hex(3)HexNAc(3)Pent(3)
C60H99N3O47S1	dHex(1)Hex(5)HexNAc(3)Sulf(1)
C64H105N3O46S0	dHex(2)Hex(3)HexNAc(3)Pent(2)
C128H210N10O90S0	Hex(4)HexNAc(5)
C65H107N3O46S0	dHex(3)Hex(3)HexNAc(3)Pent(1)
C65H107N3O47S0	dHex(2)Hex(4)HexNAc(3)Pent(1)
C62H102N4O47S1	dHex(1)Hex(4)HexNAc(4)Sulf(1)
C64H106N2O49S0	dHex(1)Hex(7)HexNAc(2)
C65H106N4O47S0	dHex(1)Hex(4)HexNAc(3)NeuAc(1)
C124H204N8O96S2	Hex(5)HexNAc(4)Sulf(1)
C64H106N2O50S0	Hex(8)HexNAc(2)
C66H108N4O47S0	dHex(1)Hex(3)HexNAc(4)Pent(2)
C65H106N4O48S0	dHex(1)Hex(4)HexNAc(3)NeuGc(1)
C67H110N4O47S0	dHex(2)Hex(3)HexNAc(4)Pent(1)
C64H105N5O47S1	dHex(1)Hex(3)HexNAc(5)Sulf(1)
C66H109N3O49S0	dHex(1)Hex(6)HexNAc(3)
C67H109N5O47S0	dHex(1)Hex(3)HexNAc(4)NeuAc(1)
C68H112N4O47S0	dHex(3)Hex(3)HexNAc(4)
C67H110N4O48S0	dHex(1)Hex(4)HexNAc(4)Pent(1)
C64H105N5O48S1	Hex(4)HexNAc(5)Sulf(1)
C66H109N3O50S0	Hex(7)HexNAc(3)
C65H106N4O50S1	dHex(1)Hex(4)HexNAc(3)NeuAc(1)Sulf(1)
C69H114N4O49S0	Hex(5)HexNAc(4)Me(2)Pent(1)
C66H108N6O48S1	Hex(3)HexNAc(6)Sulf(1)
C66H109N3O52S1	dHex(1)Hex(6)HexNAc(3)Sulf(1)
C70H115N5O49S0	dHex(1)Hex(4)HexNAc(5)
C66H107N3O53S1	dHex(1)Hex(5)HexA(1)HexNAc(3)Sulf(1)
C142H236N8O100S0	Hex(6)HexNAc(4)Me(3)
C68H112N4O51S1	dHex(2)Hex(4)HexNAc(4)Sulf(1)
C70H113N5O51S0	Hex(4)HexNAc(3)NeuAc(2)
C71H116N4O51S0	dHex(1)Hex(3)HexNAc(4)Pent(3)
C71H117N3O52S0	dHex(2)Hex(5)HexNAc(3)Pent(1)
C68H112N4O52S1	dHex(1)Hex(5)HexNAc(4)Sulf(1)
C72H118N4O51S0	dHex(2)Hex(3)HexNAc(4)Pent(2)
C71H116N4O52S0	dHex(1)Hex(5)HexNAc(3)NeuAc(1)
C66H108N6O51S2	Hex(3)HexNAc(6)Sulf(2)
C70H116N2O55S0	Hex(9)HexNAc(2)
C72H118N6O50S0	Hex(4)HexNAc(6)
C73H120N4O51S0	dHex(3)Hex(3)HexNAc(4)Pent(1)
C71H116N4O53S0	dHex(1)Hex(5)HexNAc(3)NeuGc(1)
C73H120N4O52S0	dHex(2)Hex(4)HexNAc(4)Pent(1)
C70H115N5O52S1	dHex(1)Hex(4)HexNAc(5)Sulf(1)
C72H119N3O54S0	dHex(1)Hex(7)HexNAc(3)
C73H120N4O53S0	dHex(1)Hex(5)HexNAc(4)Pent(1)
C66H107N3O56S2	dHex(1)Hex(5)HexA(1)HexNAc(3)Sulf(2)
C74H121N7O50S0	Hex(3)HexNAc(7)
C74H122N4O53S0	dHex(2)Hex(5)HexNAc(4)
C71H116N4O54S1	dHex(2)Hex(4)HexNAc(3)NeuAc(1)Sulf(1)
C68H112N4O55S2	dHex(1)Hex(5)HexNAc(4)Sulf(2)
C75H124N4O53S0	dHex(1)Hex(5)HexNAc(4)Me(2)Pent(1)
C73H119N5O54S0	Hex(5)HexNAc(4)NeuGc(1)
C72H118N6O52S1	dHex(1)Hex(3)HexNAc(6)Sulf(1)
C74H122N4O54S0	dHex(1)Hex(6)HexNAc(4)
C71H116N4O55S1	dHex(1)Hex(5)HexNAc(3)NeuAc(1)Sulf(1)
C74H122N4O55S0	Hex(7)HexNAc(4)
C71H116N4O56S1	dHex(1)Hex(5)HexNAc(3)NeuGc(1)Sulf(1)
C75H122N6O53S0	Hex(4)HexNAc(5)NeuAc(1)
C76H126N4O54S0	Hex(6)HexNAc(4)Me(3)Pent(1)
C72H119N3O57S1	dHex(1)Hex(7)HexNAc(3)Sulf(1)
C152H250N10O108S0	dHex(1)Hex(5)HexNAc(5)
C73H119N5O55S1	dHex(1)Hex(4)HexNAc(4)NeuAc(1)Sulf(1)
C74H122N4O55S1	dHex(3)Hex(4)HexNAc(4)Sulf(1)
C74H121N7O53S1	Hex(3)HexNAc(7)Sulf(1)
C76H125N5O55S0	Hex(6)HexNAc(5)
C73H119N5O56S1	Hex(5)HexNAc(4)NeuAc(1)Sulf(1)
C77H125N7O53S0	Hex(3)HexNAc(6)NeuAc(1)
C78H128N6O53S0	dHex(2)Hex(3)HexNAc(6)
C25H40N2O19S0	Hex(1)HexNAc(1)NeuGc(1)
C26H43N1O19S0	dHex(1)Hex(2)HexNAc(1)
C24H39N3O18S1	HexNAc(3)Sulf(1)
C26H43N1O20S0	Hex(3)HexNAc(1)
C23H37N1O21S1	Hex(1)HexNAc(1)Kdn(1)Sulf(1)
C27H43N3O18S0	HexNAc(2)NeuAc(1)
C26H41N1O21S0	HexNAc(1)Kdn(2)
C27H45N1O20S0	Hex(3)HexNAc(1)Me(1)
C23H36N0O23S1	Hex(2)HexA(1)Pent(1)Sulf(1)
C27H43N3O19S0	HexNAc(2)NeuGc(1)
C75H120N6O63S3	Hex(1)HexNAc(1)NeuAc(1)Sulf(1)
C28H44N2O21S0	Hex(1)HexA(1)HexNAc(2)
C26H43N1O22S1	dHex(1)Hex(2)HexNAc(1)Sulf(1)
C30H49N3O19S0	dHex(1)HexNAc(3)
C29H47N1O22S0	dHex(1)Hex(1)HexNAc(1)Kdn(1)
C30H49N3O20S0	Hex(1)HexNAc(3)
C27H43N3O21S1	HexNAc(2)NeuAc(1)Sulf(1)
C30H50N0O23S0	dHex(2)Hex(3)
C26H41N1O24S1	Hex(2)HexA(1)HexNAc(1)Sulf(1)
C30H48N0O24S0	dHex(2)Hex(2)HexA(1)
C28H46N2O22S1	dHex(1)Hex(1)HexNAc(2)Sulf(1)
C31H50N2O22S0	dHex(1)Hex(1)HexNAc(1)NeuAc(1)
C28H46N2O23S1	Hex(2)HexNAc(2)Sulf(1)
C30H50N0O25S0	Hex(5)
C32H52N4O20S0	HexNAc(4)
C30H47N3O23S0	HexNAc(1)NeuGc(2)
C31H50N2O23S0	dHex(1)Hex(1)HexNAc(1)NeuGc(1)
C32H53N1O23S0	dHex(2)Hex(2)HexNAc(1)
C31H50N2O24S0	Hex(2)HexNAc(1)NeuGc(1)
C32H53N1O24S0	dHex(1)Hex(3)HexNAc(1)
C32H51N1O25S0	dHex(1)Hex(2)HexA(1)HexNAc(1)
C30H49N3O23S1	Hex(1)HexNAc(3)Sulf(1)
C32H53N1O25S0	Hex(4)HexNAc(1)
C33H53N3O23S0	Hex(1)HexNAc(2)NeuAc(1)
C33H53N3O24S0	Hex(1)HexNAc(2)NeuGc(1)
C105H171N3O78S0	dHex(2)Hex(1)HexNAc(1)Kdn(1)
C32H53N1O27S1	dHex(1)Hex(3)HexNAc(1)Sulf(1)
C36H59N3O24S0	dHex(1)Hex(1)HexNAc(3)
C32H51N1O28S1	dHex(1)Hex(2)HexA(1)HexNAc(1)Sulf(1)
C36H59N3O25S0	Hex(2)HexNAc(3)
C33H53N3O26S1	Hex(1)HexNAc(2)NeuAc(1)Sulf(1)
C36H60N0O28S0	dHex(2)Hex(4)
C27H42N2O19S0	Hex(1)HexNAc(1)NeuAc(1)Ac(1)
C37H60N2O26S0	dHex(2)HexNAc(2)Kdn(1)
C34H56N2O27S1	dHex(1)Hex(2)HexNAc(2)Sulf(1)
C38H62N4O24S0	dHex(1)HexNAc(4)
C36H57N3O27S0	Hex(1)HexNAc(1)NeuAc(1)NeuGc(1)
C37H60N2O27S0	dHex(1)Hex(1)HexNAc(2)Kdn(1)
C36H57N3O28S0	Hex(1)HexNAc(1)NeuGc(2)
C38H59N3O27S0	Hex(1)HexNAc(1)NeuAc(2)Ac(1)
C38H61N1O29S0	dHex(2)Hex(2)HexA(1)HexNAc(1)
C36H59N3O27S1	dHex(1)Hex(1)HexNAc(3)Sulf(1)
C34H53N1O31S1	Hex(2)HexA(1)NeuAc(1)Pent(1)Sulf(1)
C39H63N3O27S0	dHex(1)Hex(1)HexNAc(2)NeuAc(1)
C38H61N1O30S0	dHex(1)Hex(3)HexA(1)HexNAc(1)
C36H59N3O28S1	Hex(2)HexNAc(3)Sulf(1)
C38H63N1O30S0	Hex(5)HexNAc(1)
C40H65N5O25S0	HexNAc(5)
C40H61N3O28S0	Hex(1)HexNAc(1)NeuAc(2)Ac(2)
C39H63N3O29S0	Hex(2)HexNAc(2)NeuGc(1)
C120H192N6O90S0	dHex(1)Hex(2)HexA(1)HexNAc(2)
C38H63N1O31S1	dHex(2)Hex(3)HexNAc(1)Sulf(1)
C41H66N4O28S0	Hex(1)HexNAc(3)NeuAc(1)
C42H69N3O28S0	dHex(2)Hex(1)HexNAc(3)
C41H66N4O29S0	Hex(1)HexNAc(3)NeuGc(1)
C39H63N3O30S1	dHex(1)Hex(1)HexNAc(2)NeuAc(1)Sulf(1)
C38H61N1O33S1	dHex(1)Hex(3)HexA(1)HexNAc(1)Sulf(1)
C42H67N3O30S0	dHex(1)Hex(1)HexA(1)HexNAc(3)
C39H63N3O31S1	Hex(2)HexNAc(2)NeuAc(1)Sulf(1)
C40H66N2O31S1	dHex(2)Hex(2)HexNAc(2)Sulf(1)
C43H70N2O31S0	dHex(2)Hex(1)HexNAc(2)Kdn(1)
C44H72N4O29S0	dHex(1)Hex(1)HexNAc(4)
C44H72N4O30S0	Hex(2)HexNAc(4)
C42H67N3O33S0	Hex(2)HexNAc(1)NeuGc(2)
C44H73N1O33S0	dHex(2)Hex(4)HexNAc(1)
C44H70N4O31S0	Hex(1)HexNAc(2)NeuAc(2)
C45H73N3O31S0	dHex(2)Hex(1)HexNAc(2)NeuAc(1)
C42H69N3O32S1	dHex(1)Hex(2)HexNAc(3)Sulf(1)
C46H75N5O29S0	dHex(1)HexNAc(5)
C45H73N3O32S0	dHex(2)Hex(1)HexNAc(2)NeuGc(1)
C46H76N2O32S0	dHex(3)Hex(2)HexNAc(2)
C42H69N3O33S1	Hex(3)HexNAc(3)Sulf(1)
C40H66N2O34S2	dHex(2)Hex(2)HexNAc(2)Sulf(2)
C45H73N3O33S0	dHex(1)Hex(2)HexNAc(2)NeuGc(1)
C47H76N4O32S0	dHex(1)Hex(1)HexNAc(3)NeuAc(1)
C138H222N6O105S0	dHex(1)Hex(3)HexA(1)HexNAc(2)
C47H76N4O33S0	dHex(1)Hex(1)HexNAc(3)NeuGc(1)
C44H70N4O34S1	Hex(1)HexNAc(2)NeuAc(2)Sulf(1)
C44H71N1O37S1	dHex(2)Hex(3)HexA(1)HexNAc(1)Sulf(1)
C47H74N4O34S0	Hex(1)HexNAc(1)NeuAc(3)
C47H76N4O34S0	Hex(2)HexNAc(3)NeuGc(1)
C45H73N3O35S1	dHex(1)Hex(2)HexNAc(2)NeuAc(1)Sulf(1)
C49H80N2O35S0	dHex(3)Hex(1)HexNAc(2)Kdn(1)
C46H76N2O36S1	dHex(2)Hex(3)HexNAc(2)Sulf(1)
C49H80N2O36S0	dHex(2)Hex(2)HexNAc(2)Kdn(1)
C46H74N2O37S1	dHex(2)Hex(2)HexA(1)HexNAc(2)Sulf(1)
C50H82N4O34S0	dHex(1)Hex(2)HexNAc(4)
C47H74N4O37S0	Hex(1)HexNAc(1)NeuGc(3)
C47H76N4O35S1	dHex(1)Hex(1)HexNAc(3)NeuAc(1)Sulf(1)
C46H74N2O38S1	dHex(1)Hex(3)HexA(1)HexNAc(2)Sulf(1)
C50H80N4O35S0	dHex(1)Hex(1)HexNAc(2)NeuAc(2)
C51H83N3O35S0	dHex(3)HexNAc(3)Kdn(1)
C47H76N4O36S1	Hex(2)HexNAc(3)NeuAc(1)Sulf(1)
C48H79N3O36S1	dHex(2)Hex(2)HexNAc(3)Sulf(1)
C52H85N5O33S0	dHex(2)HexNAc(5)
C50H80N4O36S0	Hex(2)HexNAc(2)NeuAc(2)
C51H83N3O36S0	dHex(2)Hex(2)HexNAc(2)NeuAc(1)
C48H79N3O37S1	dHex(1)Hex(3)HexNAc(3)Sulf(1)
C51H83N3O37S0	dHex(2)Hex(2)HexNAc(2)NeuGc(1)
C52H85N5O35S0	Hex(2)HexNAc(5)
C51H83N3O38S0	dHex(1)Hex(3)HexNAc(2)NeuGc(1)
C52H83N5O36S0	Hex(1)HexNAc(3)NeuAc(2)
C53H86N4O37S0	dHex(1)Hex(2)HexNAc(3)NeuAc(1)
C54H89N3O37S0	dHex(3)Hex(2)HexNAc(3)
C156H252N6O120S0	dHex(1)Hex(4)HexA(1)HexNAc(2)
C53H86N4O38S0	Hex(3)HexNAc(3)NeuAc(1)
C52H82N2O41S0	dHex(1)Hex(3)HexA(2)HexNAc(2)
C50H80N4O39S1	Hex(2)HexNAc(2)NeuAc(2)Sulf(1)
C51H83N3O39S1	dHex(2)Hex(2)HexNAc(2)NeuAc(1)Sulf(1)
C53H86N4O39S0	Hex(3)HexNAc(3)NeuGc(1)
C55H90N2O39S0	dHex(4)Hex(1)HexNAc(2)Kdn(1)
C55H90N2O40S0	dHex(3)Hex(2)HexNAc(2)Kdn(1)
C52H84N2O41S1	dHex(3)Hex(2)HexA(1)HexNAc(2)Sulf(1)
C55H89N5O38S0	Hex(2)HexNAc(4)NeuAc(1)
C56H92N4O38S0	dHex(2)Hex(2)HexNAc(4)
C52H84N2O42S1	dHex(2)Hex(3)HexA(1)HexNAc(2)Sulf(1)
C57H93N3O39S0	dHex(4)HexNAc(3)Kdn(1)
C53H84N4O42S0	Hex(2)HexNAc(1)NeuGc(3)
C56H91N1O42S0	dHex(4)Hex(1)HexNAc(1)Kdn(2)
C53H86N4O40S1	dHex(1)Hex(2)HexNAc(3)NeuAc(1)Sulf(1)
C56H90N4O40S0	dHex(1)Hex(2)HexNAc(2)NeuAc(2)
C57H93N3O40S0	dHex(3)Hex(1)HexNAc(3)Kdn(1)
C53H86N4O41S1	Hex(3)HexNAc(3)NeuAc(1)Sulf(1)
C56H90N4O41S0	Hex(3)HexNAc(2)NeuAc(2)
C53H86N4O42S1	Hex(3)HexNAc(3)NeuGc(1)Sulf(1)
C56H90N4O42S0	dHex(1)Hex(2)HexNAc(2)NeuGc(2)
C57H93N3O42S0	dHex(2)Hex(3)HexNAc(2)NeuGc(1)
C54H87N3O43S1	dHex(1)Hex(3)HexA(1)HexNAc(3)Sulf(1)
C58H93N5O41S0	Hex(2)HexNAc(3)NeuAc(2)
C59H96N4O41S0	dHex(2)Hex(2)HexNAc(3)NeuAc(1)
C60H99N3O41S0	dHex(4)Hex(2)HexNAc(3)
C58H93N5O42S0	Hex(2)HexNAc(3)NeuAc(1)NeuGc(1)
C59H96N4O42S0	dHex(2)Hex(2)HexNAc(3)NeuGc(1)
C60H99N3O42S0	dHex(3)Hex(3)HexNAc(3)
C168H270N12O129S3	dHex(1)Hex(2)HexNAc(2)NeuAc(2)Sulf(1)
C58H93N5O43S0	Hex(2)HexNAc(3)NeuGc(2)
C61H100N2O44S0	dHex(4)Hex(2)HexNAc(2)Kdn(1)
C61H99N5O42S0	dHex(1)Hex(2)HexNAc(4)NeuAc(1)
C62H102N4O42S0	dHex(3)Hex(2)HexNAc(4)
C58H91N5O46S0	Hex(1)HexNAc(1)NeuGc(4)
C63H103N3O44S0	dHex(4)Hex(1)HexNAc(3)Kdn(1)
C56H92N4O46S2	Hex(4)HexNAc(4)Sulf(2)
C63H103N3O45S0	dHex(3)Hex(2)HexNAc(3)Kdn(1)
C64H105N5O43S0	dHex(2)Hex(2)HexNAc(5)
C60H97N3O47S1	dHex(2)Hex(3)HexA(1)HexNAc(3)Sulf(1)
C60H97N3O48S1	dHex(1)Hex(4)HexA(1)HexNAc(3)Sulf(1)
C64H103N5O46S0	Hex(3)HexNAc(3)NeuAc(2)
C65H106N4O46S0	dHex(2)Hex(3)HexNAc(3)NeuAc(1)
C66H109N3O46S0	dHex(4)Hex(3)HexNAc(3)
C204H333N21O129S0	dHex(2)HexNAc(7)
C64H101N5O51S0	Hex(2)HexNAc(1)NeuGc(4)
C64H103N5O49S1	Hex(3)HexNAc(3)NeuAc(2)Sulf(1)
C70H115N5O48S0	dHex(2)Hex(3)HexNAc(5)
C67H107N5O51S0	dHex(1)Hex(2)HexNAc(2)NeuGc(3)
C66H107N3O52S1	dHex(2)Hex(4)HexA(1)HexNAc(3)Sulf(1)
C69H110N6O49S0	Hex(2)HexNAc(3)NeuAc(3)
C70H113N5O50S0	dHex(1)Hex(3)HexNAc(3)NeuAc(2)
C71H116N4O50S0	dHex(3)Hex(3)HexNAc(3)NeuAc(1)
C69H110N6O52S0	Hex(2)HexNAc(3)NeuGc(3)
C216H348N18O150S0	dHex(1)Hex(2)HexNAc(4)NeuAc(2)
C69H108N6O55S0	Hex(1)HexNAc(1)NeuGc(5)
C67H109N5O54S2	Hex(4)HexNAc(4)NeuAc(1)Sulf(2)
C67H109N5O55S2	Hex(4)HexNAc(4)NeuGc(1)Sulf(2)
C76H123N5O54S0	dHex(2)Hex(3)HexNAc(3)NeuAc(2)
C67H109N5O57S3	Hex(4)HexNAc(4)NeuAc(1)Sulf(3)
C40H66N2O28S0	dHex(2)Hex(2)HexNAc(2)
C40H66N2O29S0	dHex(1)Hex(3)HexNAc(2)
C42H69N3O29S0	dHex(1)Hex(2)HexNAc(3)
C42H69N3O30S0	Hex(3)HexNAc(3)
C40H66N2O32S1	dHex(1)Hex(3)HexNAc(2)Sulf(1)
C46H76N2O33S0	dHex(2)Hex(3)HexNAc(2)
C46H76N2O34S0	dHex(1)Hex(4)HexNAc(2)
C48H79N3O33S0	dHex(2)Hex(2)HexNAc(3)
C48H79N3O34S0	dHex(1)Hex(3)HexNAc(3)
C48H79N3O35S0	Hex(4)HexNAc(3)
C52H86N2O38S0	dHex(2)Hex(4)HexNAc(2)
C54H89N3O38S0	dHex(2)Hex(3)HexNAc(3)
C58H95N5O40S0	Hex(3)HexNAc(5)
C59H96N4O43S0	Hex(4)HexNAc(3)NeuAc(1)
C62H102N4O43S0	dHex(2)Hex(3)HexNAc(4)
C64H105N5O44S0	dHex(1)Hex(3)HexNAc(5)
C66H108N6O45S0	Hex(3)HexNAc(6)
C67H109N5O48S0	Hex(4)HexNAc(4)NeuAc(1)
C68H112N4O48S0	dHex(2)Hex(4)HexNAc(4)
C68H112N4O50S0	Hex(6)HexNAc(4)
C70H115N5O50S0	Hex(5)HexNAc(5)
C72H118N6O49S0	dHex(1)Hex(3)HexNAc(6)
C73H119N5O52S0	dHex(1)Hex(4)HexNAc(4)NeuAc(1)
C74H122N4O52S0	dHex(3)Hex(4)HexNAc(4)
C75H122N6O52S0	dHex(1)Hex(3)HexNAc(5)NeuAc(1)
C76H125N5O53S0	dHex(2)Hex(4)HexNAc(5)
C19H31N7O7S0	NQIGG
C7H6N0O2S0	Carboxyethylpyrrole
C29H19N1O7S0	Fluorescein-tyramine
C80H132N4O59S0	dHex(1)Hex(7)HexNAc(4)
C25H30N2O6S0	betaFNA
C16H32N0O0S0	Brij58
C12H24N0O0S0	Brij35
C14H20N0O0S0	Triton
C18H31N0O1S0	Tween80
C12H21N0O0S0	Tween20
C4H10N1O2S0	Tris
C18H23N3O3S1	Biotin-tyramide
C18H33N7O4S0	LRGG+dimethyl
C4H6N0O2S0	GEE
C1H0N0O-1S0	Glu->pyro-Glu+Methyl
C17H31N7O4S0	LRGG+methyl
C15H24N0O1S0	NP40
C18H16N2O8S2	IASD
C15H23N3O3S3	Biotin:Thermo-21328
C0H-1N3O0S0	AzidoF
C4H9N1O0S0	Dimethylaminoethyl
C5H6N0O3S0	Gluratylation
C4H6N0O2S0	2-hydroxyisobutyrylation
C4H8N0O2S2	DTT
C11H13N0O1S0	DYn-2
C6H8N0O4S1	Xlink:DSSO[176]
C6H10N0O1S0	MesitylOxide
C12H21N1O5S0	Xlink:DSS[259]
C1H2N0O1S0	methylol
C6H9N1O3S1	Xlink:DSSO[175]
C10H17N1O6S1	Xlink:DSSO[279]
C3H2N0O1S0	Xlink:DSSO[54]
C3H2N0O1S1	Xlink:DSSO[86]
C3H4N0O2S1	Xlink:DSSO[104]
C5H5N1O2S0	Xlink:BuUrBu[111]
C4H7N1O1S0	Xlink:BuUrBu[85]
C9H14N2O4S0	Xlink:BuUrBu[214]
C13H23N3O6S0	Xlink:BuUrBu[317]
C6H6N0O3S1	Xlink:DSSO[158]
C8H10N0O2S0	Xlink:DSS[138]
C9H12N2O3S0	Xlink:BuUrBu[196]
C6H8N2O0S2	Xlink:DTBP[172]
C4H2N0O4S0	Xlink:DST[114]
C6H6N0O2S2	Xlink:DTSSP[174]
C12H13N1O3S0	Xlink:SMCC[219]
C5H7N1O2S0	Xlink:BS2G[113]
C5H6N0O3S0	Xlink:BS2G[114]
C9H15N1O5S0	Xlink:BS2G[217]
C7H13N3O0S0	Xlink:DMP[139]
C7H10N2O0S0	Xlink:DMP[122]
C2H-2N0O0S0	glyoxalAGE
C-1H-4N0O1S-1	Met->AspSA
C-1H-2N0O-1S0	Decarboxylation
C-1H-2N-2O2S0	Aspartylurea
C-1H-1N-1O2S0	Formylasparagine
C0H-2N0O1S0	Carbonyl
C0H2N0O1S0	Pro->HAVA
C17H10N0O6S0	AFB1_Dialdehyde
C0H-4N0O2S0	Delta:H(-4)O(2)
C0H-4N0O3S0	Delta:H(-4)O(3)
C0H0N0O4S0	Delta:O(4)
C3H3N0O2S0	Delta:H(3)C(3)O(2)
C5H4N0O1S0	Delta:H(4)C(5)O(1)
C8H10N0O1S0	Delta:H(10)C(8)O(1)
C7H6N0O4S0	Delta:H(6)C(7)O(4)
C12H20N0O13S1	Hex(2)Sulf(1)
C10H16N0O8S0	Pent(2)
C13H21N1O9S0	Pent(1)HexNAc(1)
C17H28N0O13S0	Hex(1)Pent(2)Me(1)
C16H26N2O13S1	HexNAc(2)Sulf(1)
C22H36N0O17S0	Hex(1)Pent(3)Me(1)
C22H36N0O18S0	Hex(2)Pent(2)
C23H38N0O18S0	Hex(2)Pent(2)Me(1)
C30H48N0O26S0	Hex(4)HexA(1)
C31H49N1O25S0	Hex(2)HexNAc(1)Pent(1)HexA(1)
C32H51N1O26S0	Hex(3)HexNAc(1)HexA(1)
C34H56N2O26S1	Hex(1)HexNAc(2)dHex(2)Sulf(1)
C36H55N3O27S0	HexA(2)HexNAc(3)
C36H58N0O30S0	dHex(1)Hex(4)HexA(1)
C36H58N0O31S0	Hex(5)HexA(1)
C38H61N1O31S0	Hex(4)HexA(1)HexNAc(1)
C44H73N1O32S0	dHex(3)Hex(3)HexNAc(1)
C44H73N1O35S0	Hex(6)HexNAc(1)
C44H72N4O32S1	Hex(1)HexNAc(4)dHex(1)Sulf(1)
C48H77N3O35S0	dHex(1)Hex(2)HexNAc(1)NeuAc(2)
C52H86N2O37S0	dHex(3)Hex(3)HexNAc(2)
C50H82N4O36S1	dHex(2)Hex(1)HexNAc(4)Sulf(1)
C50H82N4O40S2	dHex(1)Hex(2)HexNAc(4)Sulf(2)
C54H90N0O45S0	Hex(9)
C54H89N3O41S1	dHex(2)Hex(3)HexNAc(3)Sulf(1)
C59H98N2O43S0	dHex(2)Hex(5)HexNAc(2)Me(1)
C56H92N4O44S2	dHex(2)Hex(2)HexNAc(4)Sulf(2)
C62H103N1O50S0	Hex(9)HexNAc(1)
C62H102N4O48S2	dHex(3)Hex(2)HexNAc(4)Sulf(2)
C67H109N5O49S0	Hex(4)HexNAc(4)NeuGc(1)
C69H113N3O49S0	dHex(4)Hex(3)HexNAc(2)NeuAc(1)
C69H112N6O48S0	Hex(3)HexNAc(5)NeuAc(1)
C68H113N1O55S0	Hex(10)HexNAc(1)
C70H116N2O54S0	dHex(1)Hex(8)HexNAc(2)
C72H116N6O51S0	Hex(3)HexNAc(4)NeuAc(2)
C73H119N5O51S0	dHex(2)Hex(3)HexNAc(4)NeuAc(1)
C72H118N6O51S1	dHex(2)Hex(2)HexNAc(6)Sulf(1)
C75H121N5O54S0	Hex(5)HexNAc(4)NeuAc(1)Ac(1)
C75H120N6O54S0	Hex(3)HexNAc(3)NeuAc(3)
C77H123N5O55S0	Hex(5)HexNAc(4)NeuAc(1)Ac(2)
C8H18N0O3S0	Unknown:162
C13H22N0O2S0	Unknown:210
C10H16N0O5S0	Unknown:216
C9H14N0O7S0	Unknown:234
C13H28N0O4S0	Unknown:248
C10H4N1O5S1	Unknown:250
C4H8N5O7S2	Unknown:302
C12H18N0O9S0	Unknown:306
C10H14N9O6S2	Unknown:420
C13H22N4O4S1	CIGG
C61H92N14O15S2	GNLLFLACYCIGG
'''


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='calisp',
    version='3.0.9',
    packages=setuptools.find_packages(where='src'),
    url='https://github.com/kinestetika/Calisp',
    license='MIT',
    author='Marc Strous',
    author_email='mstrous@ucalgary.ca',
    description='Isotope analysis of proteomics data',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    classifiers=['Development Status :: 4 - Beta',
                 'Environment :: Console',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.9',
                 'Topic :: Scientific/Engineering :: Bio-Informatics'],
    keywords='proteomics isotope mass-spectrometry 13C',
    project_urls={'Source': 'https://github.com/kinestetika/Calisp'},
    package_dir={'': 'src'},
    python_requires='>=3.6',
    install_requires=['numpy', 'scipy', 'pandas', 'tqdm', 'pymzml', 'pyarrow'],
    extras_require={  # Optional
        'dev': ['setuptools', 'build', 'twine'],
        'test': ['jupyter', 'matplotlib', 'jinja2'],
    },
    entry_points={  # Optional
        'console_scripts': [
            'calisp=calisp.main:main',
        ],
    }
)



import argparse
import os
from concurrent.futures import ProcessPoolExecutor

import numpy as np
import pandas as pd
import pymzml
import time
from tqdm import tqdm

from calisp import isotopic_pattern_utils
from calisp import data_store
from calisp import element_count_and_mass_utils
from calisp.peptide_spectrum_match_files import PeptideSpectrumMatchFileReader

VERSION = '3.0.9'
START_TIME = time.monotonic()
LOG_TOPICS = set()
MASS_ACCURACY = 1e-5
COMPUTE_CLUMPS = False


def format_runtime():
    runtime = time.monotonic() - START_TIME
    return f'[{int(runtime / 3600):02d}h:{int((runtime % 3600) / 60):02d}m:{int(runtime % 60):02d}s]'


def log(log_message, topic=''):
    if not topic or topic in LOG_TOPICS:
        print(f'{format_runtime()} {log_message}')


def parse_arguments():
    parser = argparse.ArgumentParser(description='src.py. (C) Marc Strous, Xiaoli Dong and Manuel Kleiner, 2018, 2021')
    parser.add_argument('--spectrum_file', required=True,  # type=argparse.FileType('r'),
                        help='[.mzML] file or folder with [.mzML] file(s).')
    parser.add_argument('--peptide_file', required=True,  # type=argparse.FileType('r'),
                        help='Search-engine-generated [.mzID] or [.target-peptide-spectrum-match] file or folder with'
                             ' these files.')
    parser.add_argument('--output_file', default='calisp-output',
                        help='The name of the folder the output gets written to. Default: calisp-output')
    parser.add_argument('--mass_accuracy', default=10, type=float,
                        help='The maximum mass difference between theoretical mass and experimental mass of a peptide')
    parser.add_argument('--bin_delimiter', default='_',
                        help='For metagenomic data, the delimiter that separates the bin ID from the protein ID '
                             '[default "_"]. Use "-" to ignore bins ID entirely.')
    parser.add_argument('--threads', default=4, type=int,
                        help='The number of (virtual) processors that calisp will use')
    parser.add_argument('--isotope', default='13C', choices=['13C', '14C', '15N', '17O', '18O', '2H', '3H', '33S',
                                                             '34S', '36S'],
                        help='The target isotope. Default: 13C')
    parser.add_argument('--compute_clumps', default=False, action='store_true',
                        help='To compute clumpiness of carbon assimilation. Only use when samples are labeled to'
                             'saturation. Estimation of clumpiness takes much additional time.')

    args = parser.parse_args()

    log(f'isotope:        {args.isotope} '
        f'[matrix{isotopic_pattern_utils.ELEMENT_ROW_INDEX, isotopic_pattern_utils.ISOTOPE_COLUMN_INDEX}]')
    log(f'peptide_file:   {args.peptide_file}')
    log(f'spectrum_file:  {args.spectrum_file}')
    log(f'output_file:    {args.output_file}')
    log(f'mass_accuracy:  {args.mass_accuracy} ppm')
    log(f'bin_delimiter:  {args.bin_delimiter}')
    log(f'threads:        {args.threads}')
    log(f'compute_clumps: {args.compute_clumps}')
    args.mass_accuracy /= 1e6
    return args


def prep_output_folder(output_folder):
    if os.path.exists(output_folder):
        log('Warning: may overwrite existing output files...')
        if os.path.isfile(output_folder):
            print(f'Expected folder at {output_folder}, found regular file, terminating now.')
            exit(1)
    else:
        os.mkdir(output_folder)


def flag_ambiguous_psms(my_psm_data: pd.DataFrame):
    for i in range(1, len(my_psm_data.index)):
        i1 = my_psm_data.index[i]
        i2 = my_psm_data.index[i-1]
        if my_psm_data.at[i1, 'psm_id'] == my_psm_data.at[i2, 'psm_id']:
            if my_psm_data.at[i1, 'psm_rank'] == my_psm_data.at[i2, 'psm_rank']:
                my_psm_data.at[i1, 'flag_psm_is_ambiguous'] = True
                my_psm_data.at[i2, 'flag_psm_is_ambiguous'] = True
            elif my_psm_data.at[i1, 'psm_rank'] > min(1, my_psm_data.at[i2, 'psm_rank']):
                my_psm_data.at[i1, 'flag_psm_is_ambiguous'] = True
            elif my_psm_data.at[i2, 'psm_rank'] > min(1, my_psm_data.at[i, 'psm_rank']):
                my_psm_data.at[i2, 'flag_psm_is_ambiguous'] = True


def set_precursor(precursors_list, my_psm_data, index):
    for precursor_dict in precursors_list:
        try:
            my_psm_data.at[index, 'psm_precursor_id'] = int(precursor_dict['precursor id'])
            my_psm_data.at[index, 'psm_precursor_mz'] = int(precursor_dict['mz'])
            return 1
        except KeyError:
            log('Warning: Unexpected definition of precursor dictionary:')
            for k in precursor_dict.keys():
                log(k, precursor_dict[k])
    return 0


def subsample_ms1_spectra(task):
    subsampled_patterns = []
    searches_done = set()
    for psm_index in task['psms'].index:
        # (1) subsample peptide isotopic patterns from precursor ms1 spectra
        precursor_id = task['psms'].at[psm_index, 'psm_precursor_id']
        if not precursor_id or task['psms'].at[psm_index, 'flag_peptide_mass_and_elements_undefined']:
            continue
        peptide_mass = task['psms'].at[psm_index, 'peptide_mass']
        precursor_i = task['ms1_spectra_hash'][precursor_id]
        initial_success = subsample_ms1_spectrum(precursor_i, psm_index, task, [], subsampled_patterns, peptide_mass,
                                                 searches_done)
        prev_success = initial_success
        # (success is a list of booleans, one boolean for each charge 1..5)
        for i in range(precursor_i+1, min(precursor_i+40, len(task['ms1_spectra_list']))):
            prev_success = subsample_ms1_spectrum(i, psm_index, task, prev_success, subsampled_patterns, peptide_mass,
                                                  searches_done)
            if not sum(prev_success):
                break
        prev_success = initial_success
        for i in range(precursor_i-1, max(0, precursor_i-40), -1):
            prev_success = subsample_ms1_spectrum(i, psm_index, task, prev_success, subsampled_patterns, peptide_mass,
                                                  searches_done)
            if not sum(prev_success):
                break
    for p in subsampled_patterns:
        # (2) assign isotopic patterns to the nearest psm:
        nearest_psm_index = p['psm_index']  # this is the psm for which the isotopic pattern was discovered
        shortest_distance = 999999
        if len(task['psms'].index) > 1:
            i_spe = task['ms1_spectra_hash'][p['pattern_precursor_id']]
            for psm_index in task['psms'].index:  # these are all the psms associated with the peptide
                i_psm = task['ms1_spectra_hash'][task['psms'].at[psm_index, 'psm_precursor_id']]
                distance_psm_to_pattern = (abs(i_psm - i_spe)+1) * (abs(task['psms'].at[psm_index, 'psm_charge']
                                                                         - p['pattern_charge']) + 1)
                if distance_psm_to_pattern < shortest_distance:
                    shortest_distance = distance_psm_to_pattern
                    nearest_psm_index = psm_index
            if nearest_psm_index != p['psm_index']:
                p['reassigned'] = True
        del p['psm_index']
        for key, value in task['psms'].loc[nearest_psm_index].to_dict().items():
            if key in p.keys():
                continue
            p[key] = value
        # (3) analyze isotopes:
        element_counts = np.array([p['C'], p['N'], p['O'], p['H'], p['S']], dtype=np.int16)
        intensities = np.empty(p['pattern_peak_count'], dtype=np.float32)
        masses = np.empty(p['pattern_peak_count'], dtype=np.float32)
        for i in range(p['pattern_peak_count']):
            intensities[i] = p[f'i{i}']
            masses[i] = p[f'm{i}']
        isotopic_pattern_utils.fit_relative_neutron_abundance(p, intensities, element_counts)
        isotopic_pattern_utils.fit_fft(p, intensities, element_counts)
        if COMPUTE_CLUMPS:
            isotopic_pattern_utils.fit_clumpy_carbon(p, intensities, element_counts)
        isotopic_pattern_utils.compute_spacing_and_irregularity(p, masses, p['pattern_charge'])
    return subsampled_patterns


def subsample_ms1_spectrum(i, psm_index, task, prev_success, subsampled_patterns, peptide_mass, searches_done):
    success = [False, False, False, False, False, False]
    if i in searches_done:
        return success
    searches_done.add(i)
    ms1_peaks = task['ms1_spectra_list'][i]
    for charge in range(1, 6):
        if len(prev_success) and not prev_success[charge]:
            continue
        expected_mz = peptide_mass / charge + element_count_and_mass_utils.PROTON_MASS
        # (1) find monoisotopic peak
        mono_isotopic_peak = []
        mip_p = 0
        for p in range(len(ms1_peaks)):
            peak = ms1_peaks[p]
            if abs(peak[0] - expected_mz) / expected_mz < MASS_ACCURACY:
                mono_isotopic_peak = peak
                mip_p = p
                break
            elif peak[0] > expected_mz:
                break
        if not len(mono_isotopic_peak):
            continue
        peak_moverz = [mono_isotopic_peak[0]]
        peak_intensities = [mono_isotopic_peak[1]]
        mass_shift = element_count_and_mass_utils.NEUTRON_MASS_SHIFT / charge

        # (2) assess presence of a peak at minus one position
        expected_mz = peak_moverz[0] - mass_shift
        peak_at_minus_one = False
        for q in range(mip_p-1, 0, -1):
            peak = ms1_peaks[q]
            if abs(peak[0] - expected_mz) / expected_mz < MASS_ACCURACY:
                peak_at_minus_one = True
            elif peak[0] < expected_mz:
                break

        # (3) find remainder of isotopic patterns
        for q in range(mip_p+1, len(ms1_peaks)):
            peak = ms1_peaks[q]
            # calculating distance from monoisotopic is slightly better than peak-to-peak
            expected_mz = peak_moverz[0] + mass_shift * len(peak_moverz)
            if abs(peak[0] - expected_mz) / expected_mz < MASS_ACCURACY:
                peak_moverz.append(peak[0])
                peak_intensities.append(peak[1])
            elif peak[0] > expected_mz:
                break

        if len(peak_moverz) >= 4:
            # normalize isotopic pattern
            peak_intensities = peak_intensities[:20]
            peak_moverz = peak_moverz[:20]
            total_intensity = sum(peak_intensities)
            peak_intensities /= total_intensity
            # assign isotopic pattern to the closest psm
            # bundle subsampled isotopic pattern as a dictionary
            subsampled_pattern = {'psm_index': psm_index,
                                   'pattern_charge': charge,
                                   'pattern_total_intensity': total_intensity,
                                   'pattern_precursor_id': task['ms1_id_list'][i],
                                   'pattern_peak_count': len(peak_intensities),
                                   'flag_pattern_is_contaminated': False,
                                   'flag_pattern_is_wobbly': False,
                                   'flag_peak_at_minus_one_pos': peak_at_minus_one,
                                   'reassigned': False}
            for p in range(len(peak_intensities)):
                subsampled_pattern[f'i{p}'] = peak_intensities[p]
                subsampled_pattern[f'm{p}'] = peak_moverz[p]
            subsampled_patterns.append(subsampled_pattern)
            success[charge] = True
    return success


def submit_for_overlap_detection(task):
    shared_patterns = task['patterns from same spectrum']
    for i in range(len(shared_patterns.index)):
        ii = shared_patterns.index[i]
        mii = set(shared_patterns.loc[ii,
                  'm0':f'm{shared_patterns.at[ii, "pattern_peak_count"] - 1}'].values)
        for j in range(i + 1, len(shared_patterns.index)):
            jj = shared_patterns.index[j]
            mjj = set(shared_patterns.loc[jj,
                      'm0':f'm{shared_patterns.at[jj, "pattern_peak_count"] - 1}'].values)
            if mjj.intersection(mii):
                return [ii, jj]
    return []


def main():
    log(f'This is calisp.py {VERSION}')
    args = parse_arguments()
    (isotopic_pattern_utils.ELEMENT_ROW_INDEX, isotopic_pattern_utils.ISOTOPE_COLUMN_INDEX) = \
        {'13C': (0, 1), '14C': (0, 2),
         '15N': (1, 1),
         '17O': (2, 1), '18O': (2, 2),
         '2H': (3, 1), '3H': (3, 2),
         '33S': (4, 1), '34S': (4, 2), '36S': (4, 4)}[args.isotope]
    global MASS_ACCURACY, COMPUTE_CLUMPS
    MASS_ACCURACY = args.mass_accuracy
    COMPUTE_CLUMPS = args.compute_clumps
    prep_output_folder(args.output_file)

    my_data_store = data_store.DataStore()
    my_data_store.set_scope(args.peptide_file, args.spectrum_file)

    log(f'Found {len(my_data_store.experiments)} peptide files and {len(my_data_store.ms_runs)} MS runs')
    for experiment_index in range(len(my_data_store.experiments)):

        # (1) read the "experiment" file with the psm and peptide definitions
        experiment_filename = my_data_store.experiments[experiment_index]
        log(f'Parsing ({experiment_index + 1}/{len(my_data_store.experiments)}) peptide file(s), '
            f'"{os.path.basename(experiment_filename)}"...')
        with PeptideSpectrumMatchFileReader(experiment_filename, bin_delimiter=args.bin_delimiter) as peptide_reader:
            psm_as_dict = []
            for psm in peptide_reader:
                psm_as_dict.append(psm)
        psm_data = pd.DataFrame(columns=data_store.DATAFRAME_COLUMNS)
        psm_data = psm_data.astype(data_store.DATAFRAME_DATATYPES)
        psm_data = pd.concat([psm_data, pd.DataFrame(psm_as_dict)], ignore_index=True)
        # psm_data.append(psm_as_dict, ignore_index=True)
        log(f'Parsed {len(psm_data.index)} PSMs from file.')
        spectrum_file_counter = 0

        ms_runs = psm_data["ms_run"].unique()
        for ms_run in ms_runs:
            try:
                ms_run_file = my_data_store.ms_run_address_book[ms_run]
            except KeyError:
                print(f'WARNING: No ms-run-file associated with ms run id "{ms_run}", skipping this file.')
                continue
            spectrum_file_counter += 1
            psm_data_of_current_ms_run = psm_data.loc[lambda df: df['ms_run'] == ms_run, :]

            # (2) sometimes, two psm's share the same ms2 spectrum. These are flagged here.
            flag_ambiguous_psms(psm_data_of_current_ms_run)

            # (3) read the ms_run file to grab currently unknown precursor ids for the known ms2 spectra as well as
            # collect all the ms1 spectra for subsampling, then: subsample.
            log(f'Now parsing ({spectrum_file_counter}/{len(ms_runs)}) spectrum files to find '
                f'{len(psm_data_of_current_ms_run.index)} MS1 precursors, "{os.path.basename(ms_run_file)}"...')
            ms1_spectra_hash = {}
            ms1_id_list = []
            ms1_spectra_list = []
            ms2_spectra_precursors = {}
            with pymzml.run.Reader(ms_run_file) as mzml_reader:
                for spectrum in mzml_reader:
                    if spectrum.ms_level == 1:
                        ms1_spectra_hash[spectrum.ID] = len(ms1_spectra_list)
                        ms1_id_list.append(spectrum.ID)
                        ms1_spectra_list.append(spectrum.peaks('centroided'))
                    elif spectrum.ms_level == 2:
                        ms2_spectra_precursors[spectrum.ID] = spectrum.selected_precursors
            psm_with_ms1_spectrum_count = 0
            for psm_index in psm_data_of_current_ms_run.index:
                try:
                    psm_with_ms1_spectrum_count += set_precursor(ms2_spectra_precursors[
                                                                 psm_data_of_current_ms_run.at[psm_index, 'psm_id']],
                                                             psm_data_of_current_ms_run, psm_index)
                except KeyError:
                    pass
            log(f'({psm_with_ms1_spectrum_count}/{len(psm_data_of_current_ms_run.index)}) PSMs have precursor MS1 spectrum')
            log(f'Now subsampling MS1 spectra for {len(psm_data_of_current_ms_run.index)} PSM patterns from '
                f'{len(ms1_spectra_list)} MS1 spectra, using {args.threads} threads for isotope analysis, '
                f'first prepping tasks for parallel execution...')

            pattern_data = pd.DataFrame(columns=data_store.DATAFRAME_COLUMNS)
            pattern_data = pattern_data.astype(data_store.DATAFRAME_DATATYPES)
            with ProcessPoolExecutor(max_workers=args.threads) as executor:
                peptides = psm_data_of_current_ms_run['peptide'].unique()
                tasks = [{'peptide': peptide,
                          'ms1_spectra_hash': ms1_spectra_hash,
                          'ms1_id_list': ms1_id_list,
                          'ms1_spectra_list': ms1_spectra_list,
                          'psms': psm_data_of_current_ms_run.loc[lambda df: df['peptide'] == peptide, :],
                          # 'psms': psm_data_of_current_ms_run[psm_data_of_current_ms_run['peptide'] == peptide],
                          } for peptide in peptides]
                patterns_reassigned_count = 0
                all_patterns = []
                for patterns in list(tqdm(executor.map(subsample_ms1_spectra, tasks, chunksize=25), total=len(tasks))):
                    for p in patterns:
                        if p['reassigned']:
                            patterns_reassigned_count += 1
                        del p['reassigned']
                    all_patterns.extend(patterns)
            pattern_data = pd.concat([pattern_data, pd.DataFrame(all_patterns)], ignore_index=True)
            # pattern_data = pattern_data.append(all_patterns, ignore_index=True)
            log(f'Subsampled and analyzed {len(pattern_data.index)} isotopic patterns, '
                f'reassigned {patterns_reassigned_count / len(pattern_data.index) * 100:.1f}% of patterns to '
                f'their nearest PSM. Now flagging overlap between patterns...')

            # (4) check for contamination (the same ms1 peaks shared by multiple peptide spectra
            with ProcessPoolExecutor(max_workers=args.threads) as executor:
                precursor_ids = pattern_data['pattern_precursor_id'].unique()
                tasks = [{'patterns from same spectrum': pattern_data[pattern_data['pattern_precursor_id'] == i]
                          } for i in precursor_ids]
                for indexes in list(tqdm(executor.map(
                        submit_for_overlap_detection, tasks, chunksize=25), total=len(tasks))):
                    for index in indexes:
                        pattern_data.at[index, 'flag_pattern_is_contaminated'] = True
            co = len(pattern_data.loc[lambda df: df["flag_pattern_is_contaminated"], :]) / len(pattern_data) * 100
            wo = len(pattern_data.loc[lambda df: df["flag_pattern_is_wobbly"], :]) / len(pattern_data) * 100
            log(f'Contamination found for {co:.1f}% of subsampled patterns. Wobbly patterns make up {wo:.1f}%.')
            ms_run_results_file = os.path.join(args.output_file,
                                               os.path.splitext(os.path.basename(ms_run_file))[0] + '.feather')
            log(f'Saving pandas dataframe with {len(pattern_data.index)} analyzed patterns to f{ms_run_results_file} '
                f'in feather format...')
            pattern_data.to_feather(ms_run_results_file)
    log(f'Done. Thanks for using calisp!')


if __name__ == "__main__":
    main()



import numpy as np
from scipy.optimize import minimize_scalar, OptimizeResult
from calisp import element_count_and_mass_utils

ELEMENT_ROW_INDEX = 0
ISOTOPE_COLUMN_INDEX = 1

MASS_SHIFT_VARIATION_TOLERANCE = 1e-5
NEUTRON_MASS_SHIFT_TOLERANCE = 0.002

ISOTOPE_MATRIX = np.array([
        [0.988943414833479, 0.011056585166521, 0.0,         0.0, 0.0,    0.0, 0.0], # C
        [0.996323567,       0.003676433,       0.0,         0.0, 0.0,    0.0, 0.0], # N
        [0.997574195,       0.00038,           0.002045805, 0.0, 0.0,    0.0, 0.0], # O
        [0.99988,           0.00012,           0.0,         0.0, 0.0,    0.0, 0.0], # H
        [0.9493,            0.0076,            0.0429,      0.0, 0.0002, 0.0, 0.0]  # S
    ], dtype=np.float32)

NATURAL_ABUNDANCES = np.array([
        [0.988943414833479, 0.011056585166521, 0.0,         0.0, 0.0,    0.0, 0.0], # C
        [0.996323567,       0.003676433,       0.0,         0.0, 0.0,    0.0, 0.0], # N
        [0.997574195,       0.00038,           0.002045805, 0.0, 0.0,    0.0, 0.0], # O
        [0.99988,           0.00012,           0.0,         0.0, 0.0,    0.0, 0.0], # H
        [0.9493,            0.0076,            0.0429,      0.0, 0.0002, 0.0, 0.0]  # S
    ], dtype=np.float32)

CLUMPY_CARBON_BOUNDS = [0.1, 0.1, 0.05, 0.0375, 0.02, 0.02, 0.02]

# VPDB standard 13C/12C = 0.0111802 in Isodat software
# https://www.webelements.com/sulfur/isotopes.html
# see also http://iupac.org/publications/pac/pdf/2003/pdf/7506x0683.pdf


def compute_relative_neutron_abundance_of_non_target_isotopes(element_counts, matrix=NATURAL_ABUNDANCES):
    rna_untargeted = np.float32(0)
    for row in range(len(matrix)):
        for column in range(len(matrix[0])):
            if row == ELEMENT_ROW_INDEX and column == ISOTOPE_COLUMN_INDEX:
                continue
            rna_untargeted += column * element_counts[row] * matrix[row][column]
    return rna_untargeted


def compute_relative_neutron_abundance(peaks, element_counts):
    rna = np.sum(peaks * range(len(peaks))) / np.sum(peaks)
    rna -= compute_relative_neutron_abundance_of_non_target_isotopes(element_counts)
    rna /= element_counts[ELEMENT_ROW_INDEX]
    rna /= ISOTOPE_COLUMN_INDEX
    return rna


def standard_ratio():
    return ISOTOPE_MATRIX[ELEMENT_ROW_INDEX][ISOTOPE_COLUMN_INDEX] / ISOTOPE_MATRIX[ELEMENT_ROW_INDEX][0]


def __fft(element_counts: np.ndarray, matrix: np.ndarray, modeled_peaks: np.ndarray):
    fft_vector_size = len(modeled_peaks)
    transforms_1 = np.empty((len(matrix), fft_vector_size), dtype=np.complex64)
    for i in range(len(matrix)):  # iterate over rows (elements)
        v = np.empty(len(matrix[0]), dtype=np.float32)
        for j in range(len(matrix[0])):  # iterate over columns (isotopes)
            v[j] = matrix[i][j]
        v_v: np.ndarray = np.fft.fft(v, n=fft_vector_size)  # norm{“backward”, “ortho”, “forward”}, optional
        for j in range(fft_vector_size):  # iterate over spectrum
            v_v[j] = pow(v_v[j], element_counts[i])
        transforms_1[i] = v_v
    transforms_2 = np.empty(fft_vector_size, dtype=np.complex64)
    for j in range(fft_vector_size):  # iterate over spectrum
        v_w = complex(1, 0)
        for i in range(len(matrix)):  # iterate over rows (elements):
            v_w *= transforms_1[i][j]
        transforms_2[j] = v_w
    v_x: np.ndarray = np.fft.ifft(transforms_2)  # inverse FFT
    total_intensity = np.float32(0)
    for j in range(fft_vector_size):  # iterate over spectrum
        total_intensity += v_x[j].real
    for j in range(fft_vector_size):  # iterate over spectrum
        modeled_peaks[j] = v_x[j].real / total_intensity


def __fft_vector_len(peak_count, element_counts: np.ndarray):
    fft_vector_len = 128
    if peak_count <= 48:
        fft_vector_len = 64
    if peak_count <= 24:
        fft_vector_len = 32
    if peak_count <= 12:
        fft_vector_len = 16
    if peak_count <= 4:
        fft_vector_len = 8
    modeled_peaks = np.zeros(fft_vector_len, dtype=np.float32)
    __fft(element_counts, ISOTOPE_MATRIX, modeled_peaks)
    while modeled_peaks[-1] > 1e-6:
        fft_vector_len *= 2
        modeled_peaks = np.zeros(fft_vector_len, dtype=np.float32)
        __fft(element_counts, ISOTOPE_MATRIX, modeled_peaks)
        if fft_vector_len >= 256:
            break
    return fft_vector_len


def __fft_fitting_function(rna, element_counts, matrix, experimental_peaks, modeled_peaks):
    matrix[ELEMENT_ROW_INDEX][ISOTOPE_COLUMN_INDEX] = rna
    matrix[ELEMENT_ROW_INDEX][0] = 1 - np.sum(matrix[ELEMENT_ROW_INDEX][1:])
    __fft(element_counts, matrix, modeled_peaks)
    return distance_ss(experimental_peaks, modeled_peaks)


def __fft_fitting_function_clumpy_carbon(rna, element_counts, matrix, experimental_peaks, modeled_peaks, i):
    matrix[ELEMENT_ROW_INDEX][i] = rna
    matrix[ELEMENT_ROW_INDEX][:i] = matrix[ELEMENT_ROW_INDEX][:i] * (1-rna) / sum(matrix[ELEMENT_ROW_INDEX][:i])
    __fft(element_counts, matrix, modeled_peaks)
    return distance_ss(experimental_peaks, modeled_peaks, i+1)


def distance_ss(peaks_a: np.ndarray, peaks_b: np.ndarray, max_peak_count=9999):  # assumes normalized spectra
    shared_peak_count = min(len(peaks_a), len(peaks_b), max_peak_count)
    if not shared_peak_count:
        return 9999

    peaks_a = peaks_a[:shared_peak_count]
    peaks_b = peaks_b[:shared_peak_count]

    total_a = np.sum(peaks_a)
    total_b = np.sum(peaks_b)
    if not total_a and not total_b:
        return 9999
    elif not total_a:
        return 9999
    elif not total_b:
        return 9999
    peaks_a = peaks_a / total_a
    peaks_b = peaks_b / total_b
    sum_of_squares = np.float32(0)
    for i in range(shared_peak_count):
        sum_of_squares += (peaks_a[i] - peaks_b[i]) ** 2
    return sum_of_squares


def fit_relative_neutron_abundance(spectrum: {}, experimental_peaks: np.ndarray, element_counts: np.ndarray):
    if not sum(element_counts):
        return {}
    rna = compute_relative_neutron_abundance(experimental_peaks, element_counts)
    ratio = rna / (1 - rna)
    spectrum['ratio_na'] = ratio
    return spectrum


def fit_fft(pattern: {}, experimental_peaks: np.ndarray, element_counts: np.ndarray,
            matrix: np.ndarray = ISOTOPE_MATRIX):
    if not sum(element_counts):
        return {}
    fft_vector_size = __fft_vector_len(len(experimental_peaks), element_counts)
    modeled_peaks = np.zeros(fft_vector_size, dtype=np.float32)
    # see https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize_scalar.html
    fit: OptimizeResult = minimize_scalar(__fft_fitting_function, bounds=(0, 0.15), method='Bounded',
                                          args=(element_counts, matrix, experimental_peaks, modeled_peaks),
                                          options={'maxiter': 40})
    pattern['ratio_fft'] = fit['x']
    pattern['error_fft'] = fit['fun']
    return pattern


def fit_clumpy_carbon(pattern: {}, experimental_peaks: np.ndarray, element_counts: np.ndarray,
                      matrix: np.ndarray = ISOTOPE_MATRIX):
    if not sum(element_counts):
        return {}
    fft_vector_size = __fft_vector_len(len(experimental_peaks), element_counts)
    matrix[ELEMENT_ROW_INDEX] = np.array([1, 0, 0, 0, 0, 0, 0], dtype=np.float32)
    index_len = min(6, len(experimental_peaks)-1)
    modeled_peaks = np.zeros(fft_vector_size, dtype=np.float32)
    for i in range(1, index_len):
        fit: OptimizeResult = minimize_scalar(__fft_fitting_function_clumpy_carbon, bounds=(0, CLUMPY_CARBON_BOUNDS[i]),
                                              method='Bounded', args=(element_counts, matrix, experimental_peaks,
                                                                      modeled_peaks, i),
                                              options={'maxiter': 40})
        pattern['error_clumpy'] = fit['fun']
    ratios = matrix[ELEMENT_ROW_INDEX][1:]*range(1, len(matrix[ELEMENT_ROW_INDEX]))
    ratios /= sum(ratios)
    for i in range(index_len):
        pattern[f'c{i + 1}'] = ratios[i]
    return pattern


def compute_spacing_and_irregularity(pattern: {}, masses, charge):
    if not len(masses):
        return
    peak_mass_spacings = np.empty(len(masses)-1, dtype=np.float32)
    for i in range(1, len(masses)):
        peak_mass_spacings[i-1] = masses[i] - masses[i-1]
    spacing_count = len(peak_mass_spacings)
    if spacing_count % 2 == 0:
        median_peak_spacing = (peak_mass_spacings[int(spacing_count/2)] +
                               peak_mass_spacings[int(spacing_count/2)-1])/2
    else:
        median_peak_spacing = peak_mass_spacings[int(spacing_count/2)]
    mass_irregularity = np.float32(0)
    for i in range(1, len(masses)):
        spacing = masses[i] - masses[i-1]
        mass_irregularity += (spacing - median_peak_spacing)**2
    median_peak_spacing *= charge
    mass_irregularity *= charge
    mass_irregularity /= len(masses)
    is_wobbly = mass_irregularity > MASS_SHIFT_VARIATION_TOLERANCE and \
        abs(median_peak_spacing - element_count_and_mass_utils.NEUTRON_MASS_SHIFT) < NEUTRON_MASS_SHIFT_TOLERANCE
    pattern['pattern_median_peak_spacing'] = median_peak_spacing
    pattern['pattern_mass_irregularity'] = mass_irregularity
    pattern['flag_pattern_is_wobbly'] = is_wobbly
    return pattern
